# -*- coding: utf-8 -*-

import logging

import numpy as np

from pyhrf import xmlio
from pyhrf.jde.samplerbase import *
from pyhrf.jde.beta import *


logger = logging.getLogger(__name__)


##########################################################
# Gamma/Gaussian mixture models
##########################################################
class InhomogeneousNRLSampler(xmlio.XmlInitable, GibbsSamplerVariable):
    """Class handling the Gibbs sampling of Neural Response Levels according to:

    - Makni, S., Ciuciu, P., Idier, J., & Poline, J.-B. (2005). Joint detection-estimation of brain activity in
      functional MRI: a Multichannel Deconvolution solution. IEEE Transactions on Signal Processing, 53(9), 3488–3502.
      https://doi.org/10.1109/TSP.2005.853303

    Inherits the abstract class C{GibbsSamplerVariable}.
    #TODO : comment attributes
    """

    # parameters specifications :
    P_SAMPLE_LABELS = 'sampleLabels'
    P_LABELS_INI = 'labelsIni'
    P_BETA = 'beta'
    P_LABELS_COLORS = 'labelsColors'
    P_TRUE_LABELS = 'trueLabels'
    P_SAMPLE_FLAG = 'sampleFlag'
    P_VAL_INI = 'initialValue'

    # parameters definitions and default values :
    defaultParameters = {
        P_SAMPLE_FLAG: 1,
        P_VAL_INI: None,
        P_BETA: 0.4,
        P_SAMPLE_LABELS: 1,
        P_LABELS_INI: None,
        P_LABELS_COLORS: np.array([0.0, 0.0], dtype=float),
    }

    # other class attributes
    L_CI = 0
    L_CA = 1

    def __init__(self, parameters=None, xmlHandler=None, xmlLabel=None,
                 xmlComment=None):

        # TODO : comment
        xmlio.XmlInitable.__init__(self)
        sampleFlag = self.parameters[self.P_SAMPLE_FLAG]
        valIni = self.parameters[self.P_VAL_INI]
        GibbsSamplerVariable.__init__(self, 'nrl', valIni=valIni,
                                      sampleFlag=sampleFlag)

        # instance variables affectation from parameters :
        self.sampleLabelsFlag = self.parameters[self.P_SAMPLE_LABELS]
        self.labels = self.parameters[self.P_LABELS_INI]
        if self.parameters.has_key(self.P_TRUE_LABELS):
            self.trueLabels = params[self.P_TRUE_LABELS]
        else:
            self.trueLabels = None
        self.beta = self.parameters[self.P_BETA]
        self.constJ = 1

    def linkToData(self, dataInput):
        self.dataInput = dataInput
        self.nbConditions = self.dataInput.nbConditions
        self.nbVox = self.dataInput.nbVoxels
        self.ny = self.dataInput.ny
        self.cardCI = np.zeros(self.nbConditions, dtype=int)
        self.cardCA = np.zeros(self.nbConditions, dtype=int)
        self.voxIdxCI = range(self.nbConditions)
        self.voxIdxCA = range(self.nbConditions)

        self.cumulLabels = np.zeros((self.nbConditions, self.nbVox), dtype=int)

        self.meanClassAApost = range(self.nbConditions)
        self.meanClassIApost = range(self.nbConditions)
        self.varXjhtQjeji = range(self.nbConditions)

        # some allocations :
        for j in xrange(self.nbConditions):
            self.meanClassAApost[j] = numpy.empty((self.nbVox))
            self.meanClassIApost[j] = numpy.empty((self.nbVox))
            self.varXjhtQjeji[j] = numpy.empty((self.nbVox))

    def checkAndSetInitValue(self, variables):
        # Generating default labels if necessary :
        if self.labels == None:  # if no initial labels specified
            if not self.sampleLabelsFlag and self.dataInput.simulData != None:
                self.labels = self.dataInput.simulData.nrls.labels
            else:
                self.labels = np.zeros(
                    (self.nbConditions, self.nbVox), dtype=int)
                nbVoxInClassI = self.nbVox / 2
                # Uniform dispatching :
                for j in xrange(self.nbConditions):
                    self.labels[j, 0:nbVoxInClassI] = self.L_CI
                    self.labels[j, nbVoxInClassI:self.nbVox] = self.L_CA
                    self.labels[j, :] = np.random.permutation(
                        self.labels[j, :])

        self.countLabels()

        if self.currentValue == None:
            if not self.sampleFlag and self.dataInput.simulData != None:
                self.currentValue = self.dataInput.simulData.nrls.data
            else:
                nrlsIni = np.zeros(
                    (self.nbConditions, self.nbVox), dtype=float)
                # Init Nrls according to classes definitions :
                smplMixtP = variables[self.samplerEngine.I_MIXT_PARAM]
                # ensure that mixture parameters are correctly set
                smplMixtP.checkAndSetInitValue(variables)

                var_classI = smplMixtP.currentValue[
                    GamGaussMixtureParamsSampler.I_VAR_CI]
                shape_classA = smplMixtP.currentValue[
                    GamGaussMixtureParamsSampler.I_SHAPE_CA]
                scale_classA = smplMixtP.currentValue[
                    GamGaussMixtureParamsSampler.I_SCALE_CA]
                for j in xrange(self.nbConditions):
                    nrlsIni[j, self.voxIdxCI[j]] = np.random.randn(
                        self.cardCI[j]) * var_classI[j] ** 0.5
                    nrlsIni[j, self.voxIdxCA[j]] = np.random.gamma(
                        shape_classA[j], scale_classA[j], self.cardCA[j])
                self.currentValue = nrlsIni

    def countLabels(self):
        for j in xrange(self.nbConditions):
            self.voxIdxCI[j] = where(self.labels[j, :] == self.L_CI)[0]
            self.voxIdxCA[j] = where(self.labels[j, :] == self.L_CA)[0]
            self.cardCI[j] = len(self.voxIdxCI[j])
            self.cardCA[j] = len(self.voxIdxCA[j])
            logger.debug('Nb vox in CI for cond %d : %d', j, self.cardCI[j])
            logger.debug('Nb vox in CI for cond %d : %d', j, self.cardCA[j])

    def samplingWarmUp(self, variables):
        """
        #TODO : comment
        """
        # print 'NRLs warming up ...'

        # Compute precalculations :
        smplHRF = self.get_variable('hrf')
        smplHRF.checkAndSetInitValue(variables)
        self.normHRF = smplHRF.norm
        self.varYtilde = np.zeros((self.ny, self.nbVox), dtype=float)
        self.computeVarYTilde(smplHRF.varXh)
        self.varXhtQXh_allvox = empty(
            (self.nbConditions, self.nbVox), dtype=float)
        self.sumRmatXhtQXh = empty(
            (self.nbConditions, self.nbVox), dtype=float)
        self.varClassIApost = empty(
            (self.nbConditions, self.nbVox), dtype=float)
        self.varClassAApost = empty(
            (self.nbConditions, self.nbVox), dtype=float)
        self.sigClassIApost = empty(
            (self.nbConditions, self.nbVox), dtype=float)
        self.sigClassAApost = empty(
            (self.nbConditions, self.nbVox), dtype=float)

    def computeVarYTilde(self, varXh):
        for i in xrange(self.nbVox):
            repNRLi = repmat(self.currentValue[:, i], self.ny, 1)
            aijXjh = repNRLi * varXh
            self.varYtilde[:, i] = self.dataInput.varMBY[
                :, i] - aijXjh.sum(axis=1)

    def sampleNextAlt(self, variables):
        varXh = self.get_variable('hrf').varXh
        self.computeVarYTilde(varXh)

    def computeVariablesApost(self, varCI, shapeCA, scaleCA, rb, varXh,
                              varLambda):

        self.varXhtQ = dot(varXh.transpose(), self.dataInput.delta)
        varXhtQXh = diag(dot(self.varXhtQ, varXh))
        rmatVarXhtQXh = repeat(varXhtQXh, self.nbVox).reshape(varXhtQXh.size,
                                                              self.nbVox)

        numpy.divide(rmatVarXhtQXh, repmat(rb, self.nbConditions, 1),
                     self.varXhtQXh_allvox)

        rmatInvVarCI = repeat(1.0 / varCI, self.nbVox).reshape(self.nbConditions,
                                                               self.nbVox)

        numpy.power(numpy.add(rmatInvVarCI, self.varXhtQXh_allvox,
                              self.sumRmatXhtQXh), -1, self.varClassIApost)
        numpy.power(self.varXhtQXh_allvox, -1, self.varClassAApost)

        numpy.square(self.varClassIApost, self.sigClassIApost)
        numpy.square(self.varClassAApost, self.sigClassAApost)

#        rmatMCAVCA = repeat(meanCA/varCA, self.nbVox).reshape(self.nbConditions,
#                                                              self.nbVox)
        numpy.multiply(- scaleCA, self.varClassAApost, self.interMeanClassA)

        if self.beta <= 0:
            self.ratioLambdaI_A = ( (1 - varLambda) * numpy.gamma()(varCA ** 0.5) ) \
                / (varLambda * (varCI ** 0.5))
        else:
            self.ratioLambdaI_A = (varCA / varCI) ** 0.5

        logger.debug('ratioLambdaI_A :')
        logger.debug(self.ratioLambdaI_A)

    def computeMeanClassApost(self, j, nrls, varXhj, rb):

        logger.debug('self.varXhtQ[%d,:] :', j)
        logger.debug(self.varXhtQ[j, :])

        for i in xrange(self.nbVox):
            eji = self.varYtilde[:, i] + nrls[j, i] * varXhj
            self.varXjhtQjeji[i] = dot(self.varXhtQ[j, :], eji) / rb[i]
        numpy.add(self.interMeanClassA[j, :],
                  multiply(self.varClassAApost[j, :], self.varXjhtQjeji),
                  self.meanClassAApost[j])
        logger.debug('meanClassAApost cond %d :', j)
        logger.debug(self.meanClassAApost[j])

        multiply(self.varClassIApost[j, :],
                 self.varXjhtQjeji,
                 self.meanClassIApost[j])
        logger.debug('meanClassIApost cond %d :', j)
        logger.debug(self.meanClassIApost[j])

    def sampleNextInternal(self, variables):
        # TODO : comment

        varCI = variables[self.samplerEngine.I_MIXT_PARAM].currentValue[
            BiGaussMixtureParamsSampler.I_VAR_CI]
        varCA = variables[self.samplerEngine.I_MIXT_PARAM].currentValue[
            BiGaussMixtureParamsSampler.I_VAR_CA]
        meanCA = variables[self.samplerEngine.I_MIXT_PARAM].currentValue[
            BiGaussMixtureParamsSampler.I_MEAN_CA]
        rb = variables[self.samplerEngine.I_NOISE_VAR].currentValue
        varXh = self.get_variable('hrf').varXh
        varLambda = variables[
            self.samplerEngine.I_WEIGHTING_PROBA].currentValue

        self.computeVarYTilde(varXh)
        self.computeVariablesApost(varCI, varCA, meanCA, rb, varXh, varLambda)

        self.labelsSamples = np.random.rand(self.nbConditions, self.nbVox)
        nrlsSamples = np.random.randn(self.nbConditions, self.nbVox)

        for j in xrange(self.nbConditions):
            self.computeMeanClassApost(j, self.currentValue, varXh[:, j], rb)
            if self.sampleLabelsFlag:
                logger.info('Sampling labels ...')
                self.sampleLabels(j, varCI, varCA, meanCA)
                logger.info('Sampling labels done!')

            sigApost = np.zeros(self.nbVox, dtype=float)

            putmask(sigApost, self.labels[j, :], self.sigClassAApost[j, :])
            putmask(sigApost, 1 - self.labels[j, :], self.sigClassIApost[j, :])

            meanApost = np.zeros(self.nbVox, dtype=float)
            putmask(meanApost, self.labels[j, :], self.meanClassAApost[j])
            putmask(meanApost, 1 - self.labels[j, :], self.meanClassIApost[j])

            self.currentValue[j, :] = multiply(nrlsSamples[j, :], sigApost) + \
                meanApost
            logger.debug('All nrl cond %d:', j)
            logger.debug(self.currentValue[j, :])
            logger.info('nrl cond %d = %1.3f(%1.3f)', j,
                        self.currentValue[j, :].mean(),
                        self.currentValue[j, :].std())
            logger.debug('All nrl activ cond %d:', j)
            logger.debug(self.currentValue[j, self.voxIdxCA[j]])

            logger.info('nrl activ cond %d = %1.3f(%1.3f)', j,
                        self.currentValue[j, self.voxIdxCA[j]].mean(),
                        self.currentValue[j, self.voxIdxCA[j]].std())
            logger.debug('All nrl inactiv cond %d:', j)
            logger.debug(self.currentValue[j, self.voxIdxCI[j]])

            logger.info('nrl inactiv cond %d = %1.3f(%1.3f)', j,
                        self.currentValue[j, self.voxIdxCI[j]].mean(),
                        self.currentValue[j, self.voxIdxCI[j]].std())

            # TODO : may be improved by not completely updating varYtilde ...
            self.computeVarYTilde(varXh)

        self.countLabels()
        self.normHRF = self.get_variable('hrf').norm

    def sampleLabels(self, cond, varCI, varCA, meanCA):
        # print 'sampling labels ...'

        fracLambdaTilde = self.ratioLambdaI_A[cond] *                     \
            (self.sigClassIApost[cond, :]
             / self.sigClassAApost[cond, :] ) *            \
            np.exp(0.5 * (self.meanClassIApost[cond] ** 2
                          / self.varClassIApost[cond, :]
                          - self.meanClassAApost[cond] ** 2
                          / self.varClassAApost[cond, :]
                          + meanCA[cond] ** 2 / varCA[cond]))
        if self.beta > 0:
            # TODO : a paralleliser : possible ??
            for i in xrange(self.nbVox):
                fracLambdaTilde[i] *= np.exp(-self.beta * (self.calcEnergy(i, self.L_CI, cond) -
                                                           self.calcEnergy(i, self.L_CA, cond)))

        varLambdaApost = 1 / (1 + fracLambdaTilde)

        self.labels[cond, :] = np.array(
            self.labelsSamples[cond, :] <= varLambdaApost, dtype=int)

    def calcEnergy(self, voxIdx, label, cond):

        neighboursMask = self.dataInput.voxelMapping.getNeighboursIndexes(
            voxIdx)
        spCorr = np.sum(2 * (self.labels[cond, neighboursMask] == label) - 1.0)

        return -self.constJ * spCorr  # -self.labelsColors[label]

    def finalizeSampling(self):
        self.finalLabels = np.zeros((self.nbConditions, self.nbVox), dtype=int)
        for j in xrange(self.nbConditions):
            self.finalLabels[j, where(self.meanLabels[j, :] >= 0.5)] = 1

        if self.trueLabels != None:
            for j in xrange(self.nbConditions):
                deltaLab = self.finalLabels[j, :] - self.trueLabels[j, :]
                # false negative
                self.finalLabels[j, where(deltaLab == -1)] = -1
                self.finalLabels[j, where(deltaLab == 1)] = 2  # false positive

        smplHRF = self.samplerEngine.get_variable('hrf')

        # Correct sign ambiguity :
        if smplHRF.detectSignError():
            self.finalValue *= -1

        # Correct hrf*nrl scale ambiguity :
        scaleF = smplHRF.getScaleFactor()
        # Use HRF amplitude :
        self.finalValueScaleCorr = self.finalValue * scaleF

        # clean memory of temporary variables :
        del self.varXhtQXh_allvox
        del self.sumRmatXhtQXh
        del self.varClassIApost
        del self.varClassAApost
        del self.sigClassIApost
        del self.sigClassAApost
        del self.interMeanClassA
        del self.meanClassAApost
        del self.meanClassIApost
        del self.varXjhtQjeji

    def computeMean(self):
        # print 'computing nrls mean ...'

        self.cumul += self.currentValue  # *self.normHRF
        self.cumulLabels += self.labels

        self.nbItMean += 1
        self.mean = self.cumul / self.nbItMean
        self.meanLabels = self.cumulLabels / (self.nbItMean + 0.0)


class GamGaussMixtureParamsSampler(GibbsSamplerVariable):
    """
    #TODO : comment

    """

    I_MEAN_CA = 0
    I_VAR_CA = 1
    I_VAR_CI = 2
    NB_PARAMS = 3
    PARAMS_NAMES = ['Shape_Activ', 'Scale_Activ', 'Var_Inactiv']

    P_VAL_INI = 'initialValue'
    P_SAMPLE_FLAG = 'sampleFlag'

    P_SHAPE_CA_PR_MEAN = 'shapeCAPrMean'
    P_SCALE_CA_PR_ALPHA = 'scaleCAPrAlpha'
    P_SCALE_CA_PR_BETA = 'scaleCAPrBeta'

    P_VAR_CI_PR_ALPHA = 'varCIPrAlpha'
    P_VAR_CI_PR_BETA = 'varCIPrBeta'

    defaultParameters = {
        P_VAL_INI: None,
        P_SAMPLE_FLAG: 1,
        P_SHAPE_CA_PR_MEAN: 10.,
        P_SCALE_CA_PR_ALPHA: 2.5,
        P_SCALE_CA_PR_BETA: 1.5,
        P_VAR_CI_PR_ALPHA: 2.5,
        P_VAR_CI_PR_BETA: .5,
    }

    def __init__(self, parameters=None, xmlHandler=None,
                 xmlLabel=None, xmlComment=None):
        """
        #TODO : comment
        """
        xmlio.XMLParamDrivenClass.__init__(
            self, parameters, xmlHandler, xmlLabel, xmlComment)
        sampleFlag = self.parameters[self.P_SAMPLE_FLAG]
        valIni = self.parameters[self.P_VAL_INI]

        # get values for priors :
        self.varCIPrAlpha = self.parameters[self.P_VAR_CI_PR_ALPHA]
        self.varCIPrBeta = self.parameters[self.P_VAR_CI_PR_BETA]
        self.scaleCAPrAlpha = self.parameters[self.P_SCALE_CA_PR_ALPHA]
        self.scaleCAPrBeta = self.parameters[self.P_SCALE_CA_PR_BETA]

        self.shapeCAPrMean = self.parameters[self.P_SHAPE_CA_PR_MEAN]

        GibbsSamplerVariable.__init__(
            self, 'mixtParams', valIni=valIni, sampleFlag=sampleFlag)

    def linkToData(self, dataInput):
        self.dataInput = dataInput
        self.nbConditions = self.dataInput.nbConditions
        self.nbVox = self.dataInput.nbVoxels
        self.ny = self.dataInput.ny
        self.nbColX = self.dataInput.nbColX

        self.nrlCI = range(self.nbConditions)
        self.nrlCA = range(self.nbConditions)

    def checkAndSetInitValue(self, variables):

        if self.currentValue == None:
            self.currentValue = np.zeros(
                (self.NB_PARAMS, self.nbConditions), dtype=float)
            if not self.sampleFlag and self.dataInput.simulData != None:
                mixtures = self.dataInput.simulData.nrls.getMixture()
                itemsCond = mixtures.items()
                assert len(itemsCond) == self.nbConditions
                shapeCA = np.zeros(self.nbConditions, dtype=float)
                scaleCA = np.zeros(self.nbConditions, dtype=float)
                varCI = np.zeros(self.nbConditions, dtype=float)
                for cn, mixt in mixtures.iteritems():
                    genActiv = mixt.generators['activ']
                    genInactiv = mixt.generators.generators['inactiv']
                    indCond = self.dataInput.simulData.nrls.condIds[cn]
                    meanCA = genActiv.mean
                    varCA = genActiv.std ** 2
                    shapeCA[indCond] = meanCA ** 2 / (varCA + 0.)
                    scaleCA[indCond] = varCA / (meanCA + 0.)
                    varCI[indCond] = genInactiv.std ** 2
                self.currentValue[self.I_SHAPE_CA] = shapeCA
                self.currentValue[self.I_SCALE_CA] = scaleCA
                self.currentValue[self.I_VAR_CI] = varCI
        else:
            # TODO : put constants in default init values ...
            self.currentValue[self.I_SHAPE_CA] = np.zeros(
                self.nbConditions) + 10.0
            self.currentValue[self.I_SCALE_CA] = np.zeros(
                self.nbConditions) + 5.0
            self.currentValue[self.I_VAR_CI] = np.zeros(self.nbConditions) + 1.

    def sampleNextInternal(self, variables):
        # TODO : comment

        # print '- Sampling Mixt params ...'

        nrlsSmpl = variables[self.samplerEngine.I_NRLS]
        cardCA = nrlsSmpl.cardCA
        cardCI = nrlsSmpl.cardCI

        for j in xrange(self.nbConditions):
            self.nrlCI[j] = nrlsSmpl.currentValue[j, nrlsSmpl.voxIdxCI[j]]
            self.nrlCA[j] = nrlsSmpl.currentValue[j, nrlsSmpl.voxIdxCA[j]]

        self.currentValue = np.zeros((self.NB_PARAMS, self.nbConditions),
                                     dtype=float)

        for j in xrange(self.nbConditions):
            if cardCI[j] > 0:
                nu0j = .5 * dot(self.nrlCI[j], self.nrlCI[j])
                varCIj = 1.0 / np.random.gamma(.5 * cardCI[j] + self.varCIPrAlpha,
                                               1 / (nu0j + self.varCIPrBeta))
            else:
                varCIj = 1.0 / \
                    np.random.gamma(self.varCIPrAlpha, 1 / self.varCIPrBeta)

            self.currentValue[self.I_VAR_CI, j] = varCIj
#            self.currentValue[self.I_VAR_CI, j] = 0.0000001

            # sampling of shapeCA and scaleCA to be completed

#            invVarLikelihood = cardCA[j]/varCAj

            meanCAVarAPost = 1 / (invVarLikelihood + 1 / self.meanCAPrVar)
# print 'meanCAVarAPost = 1/(invVarLikelihood + 1/self.meanCAPrVar) :'
# print '%f = 1/(%f + 1/%f)'
# %(meanCAVarAPost,invVarLikelihood,self.meanCAPrVar)
            rPrMV = self.meanCAPrMean / self.meanCAPrVar
            meanCAMeanAPost = meanCAVarAPost * \
                (eta1j * invVarLikelihood + rPrMV)

# print 'meanCAMeanAPost = meanCAVarAPost * (eta1j*invVarLikelihood + rPrMV) :'
# print '%f = %f *(%f*%f + %f)' %(meanCAMeanAPost,meanCAVarAPost,eta1j,invVarLikelihood,rPrMV)
# print 'meanCAMeanAPost :', meanCAMeanAPost
            shapeCAj = np.random.normal(meanCAMeanAPost, meanCAVarAPost ** 0.5)

            # absolute(meanCAj)
            self.currentValue[self.I_SHAPE_CA, j] = shapeCAj
            self.currentValue[self.I_SCALE_CA, j] = scaleCAj

            logger.info(
                'varCI,%d = %f', j, self.currentValue[self.I_VAR_CI, j])
            logger.info(
                'meanCA,%d = %f', j, self.currentValue[self.I_SHAPE_CA, j])
            logger.info(
                'varCA,%d = %f', j, self.currentValue[self.I_SCALE_CA, j])



# -*- coding: utf-8 -*-

import unittest
import tempfile
import shutil
import logging

import numpy as np

import pyhrf
import pyhrf.tools as tools

from pyhrf import FmriData
from pyhrf.ui.treatment import FMRITreatment
from pyhrf.jde.asl import simulate_asl
from pyhrf.ui.vb_jde_analyser_asl_fast import JDEVEMAnalyser


logger = logging.getLogger(__name__)


class VEMASLTest(unittest.TestCase):

    def setUp(self):
        np.random.seed(8652761)

        tmpDir = tempfile.mkdtemp(prefix='pyhrf_tests',
                                  dir=pyhrf.cfg['global']['tmp_path'])
        self.tmp_dir = tmpDir
        self.clean_tmp = True
        simu = simulate_asl(self.tmp_dir, spatial_size='random_small')
        self.data_simu = FmriData.from_simulation_dict(simu)

    def tearDown(self):
        if self.clean_tmp:
            logger.info('Remove tmp dir %s', self.tmp_dir)
            shutil.rmtree(self.tmp_dir)
        else:
            logger.info('Keep tmp dir %s', self.tmp_dir)

    def test_jdevemanalyser(self):
        """ Test BOLD VEM sampler on small simulation with small
        nb of iterations. Estimation accuracy is not tested.
        """
        jde_vem_analyser = JDEVEMAnalyser(beta=.8, dt=.5, hrfDuration=25.,
                                          nItMax=2, nItMin=2, fast=True,
                                          PLOT=False,
                                          constrained=True)
        tjde_vem = FMRITreatment(fmri_data=self.data_simu,
                                 analyser=jde_vem_analyser,
                                 output_dir=None)
        tjde_vem.run()



# -*- coding: utf-8 -*-

import os.path as op
import unittest
import shutil
import math
import logging

import numpy as np

from numpy.testing import (assert_array_equal, assert_array_almost_equal,
                           assert_almost_equal, assert_array_less, assert_equal)

import pyhrf
import pyhrf.graph as pgraph
import pyhrf.sandbox.parcellation as pm
import pyhrf.boldsynth.scenarios as simu

from pyhrf.tools import assert_file_exists
from pyhrf.ndarray import expand_array_in_mask, xndarray, MRI3Daxes
from pyhrf import Condition
from pyhrf.tools import Pipeline
from pyhrf.core import FmriData
from pyhrf.parcellation import parcellation_dist
from pyhrf.sandbox.parcellation import loglikelihood_computation

# launch all the tests in here:
#  pyhrf_maketests -v test_sandbox_parcellation
# to see real data:
#  anatomist cortex_occipital_* hrf_territory
#  anatomist cortex_occipital_*


logger = logging.getLogger(__name__)


class StatTest(unittest.TestCase):

    def setUp(self):
        pass

    def test_norm_bc(self):
        a = np.array([[1., 2.],
                      [2., 4.],
                      [5., 10.]])
        b = np.array([.5, .5])

        expected_norm = np.array([np.linalg.norm(x - b) ** 2 for x in a])
        assert_almost_equal(pm.norm2_bc(a, b), expected_norm)

    def test_gmm_known_weights_simu_1D(self):
        """
        Test biGMM fit with known post weights, from biGMM samples (no noise)
        1D case.
        """
        plot = False
        # pyhrf.verbose.set_verbosity(0)
        pyhrf.logger.setLevel(logging.WARNING)
        np.random.seed(2354)  # make reproducible results

        # biGMM parameters:
        mu0 = np.array([2., 5., 10.])
        v0 = 3.

        mu1 = np.array([12., 15., 20.])
        v1 = 3.

        l = .3  # weight for component 1

        # generate samples from biGMM:
        nsamples = 10000
        samples = np.random.randn(nsamples) * v0 ** .5 + mu0[:, np.newaxis]
        card_c1 = int(l * nsamples)
        samples[:, :card_c1] = np.random.randn(card_c1) * v1 ** .5 + \
            mu1[:, np.newaxis]
        samples = samples.T

        # compute sample-specific posterior weights:
        nfeats = len(mu0)
        print nfeats
        post_weights = 1 / (1 + (1 - l) / l * (v0 / v1) ** (-nfeats / 2.) *
                            np.exp(-(pm.norm2_bc(samples, mu0) / v0 -
                                     pm.norm2_bc(samples, mu1) / v1) / 2.))

        # fit GMM knowing post weights:
        mu0_est, mu1_est, v0_est, v1_est, l_est, llh = \
            pm.informedGMM_MV(samples, post_weights)

        # Some outputs:
        mus = np.array([[mu0, mu1], [mu0_est, mu1_est]]).transpose()
        vs = np.array([[v0, v1], [v0_est, v1_est]]).T
        ls = np.array([[1 - l, l], [1 - l_est, l_est]]).T

        from pyhrf.tools import get_2Dtable_string
        logger.info('means:')
        for f in xrange(samples.shape[1]):
            logger.info(get_2Dtable_string(mus[f, :, :], ['C0', 'C1'],
                                           ['true', 'estim']))
        logger.info('vars:')
        logger.info(get_2Dtable_string(vs, ['C0', 'C1'], ['true', 'estim']))
        logger.info('weights:')
        logger.info(get_2Dtable_string(ls, ['C0', 'C1'], ['true', 'estim']))

        if plot:
            import matplotlib.pyplot as plt
            from pyhrf.plot import plot_gaussian_mixture
            f = 1  # check this feature

            plot_gaussian_mixture(np.array([[mu0[f], mu1[f]], [v0, v1]]),
                                  props=[1 - l, l], color='r')

            plot_gaussian_mixture(np.array([[mu0_est[f], mu1_est[f]],
                                            [v0_est, v1_est]]),
                                  props=[1 - l_est, l_est], color='b')

            plt.hist(samples[:, f], color='g', normed=True)
            plt.show()

        assert_array_almost_equal([mu0_est, mu1_est], [mu0, mu1], decimal=1)
        assert_array_almost_equal([v0_est, v1_est], [v0, v1], decimal=1)
        assert_array_almost_equal([1 - l_est, l_est], [1 - l, l], decimal=1)

    def test_gmm_known_weights_difvars_noise(self):
        """
        Test biGMM fit with known post weights, from biGMM samples (no noise)
        1D case.
        """
        plot = False
        # pyhrf.verbose.set_verbosity(0)
        pyhrf.logger.setLevel(logging.WARNING)
        np.random.seed(2354)  # make reproducible results

        # biGMM parameters:
        mu0 = np.array([2., 5., 10.])
        v0 = 3.

        mu1 = np.array([12., 15., 20.])
        v1 = v0 / 50.

        l = .3  # weight for component 1
        v_noise = 1.5

        # generate samples from biGMM:
        nsamples = 10000
        samples = np.random.randn(nsamples) * v0 ** .5 + mu0[:, np.newaxis]
        card_c1 = int(l * nsamples)
        samples[:, :card_c1] = np.random.randn(card_c1) * v1 ** .5 + \
            mu1[:, np.newaxis]
        samples = samples.T
        noise = np.random.randn(samples.shape[0], samples.shape[1]) * v_noise
        # print 'noise shape = ', str(noise.shape)
        noisy_samples = samples + noise

        # compute sample-specific posterior weights:
        nfeats = len(mu0)
        print nfeats
        post_weights = 1 / (1 + (1 - l) / l * (v0 / v1) ** (-nfeats / 2.) *
                            np.exp(-(pm.norm2_bc(samples, mu0) / v0 -
                                     pm.norm2_bc(samples, mu1) / v1) / 2.))

        # fit GMM knowing post weights:
        mu0_est, mu1_est, v0_est, v1_est, l_est, llh = \
            pm.informedGMM_MV(noisy_samples, post_weights)

        # Some outputs:
        mus = np.array([[mu0, mu1], [mu0_est, mu1_est]]).transpose()
        vs = np.array([[v0, v1], [v0_est, v1_est]]).T
        ls = np.array([[1 - l, l], [1 - l_est, l_est]]).T

        from pyhrf.tools import get_2Dtable_string
        logger.info('means:')
        for f in xrange(samples.shape[1]):
            logger.info(get_2Dtable_string(mus[f, :, :], ['C0', 'C1'],
                                           ['true', 'estim']))
        logger.info('vars:')
        logger.info(get_2Dtable_string(vs, ['C0', 'C1'], ['true', 'estim']))
        logger.info('weights:')
        logger.info(get_2Dtable_string(ls, ['C0', 'C1'], ['true', 'estim']))

        if plot:
            import matplotlib.pyplot as plt
            from pyhrf.plot import plot_gaussian_mixture
            f = 1  # check this feature

            plot_gaussian_mixture(np.array([[mu0[f], mu1[f]], [v0, v1]]),
                                  props=[1 - l, l], color='r')

            plot_gaussian_mixture(np.array([[mu0_est[f], mu1_est[f]],
                                            [v0_est, v1_est]]),
                                  props=[1 - l_est, l_est], color='b')

            plt.hist(samples[:, f], color='g', normed=True)
            plt.show()

        assert_array_almost_equal([mu0_est, mu1_est], [mu0, mu1], decimal=1)
        assert_array_almost_equal([v0_est, v1_est], [v0, v1], decimal=1)
        assert_array_almost_equal([1 - l_est, l_est], [1 - l, l], decimal=1)

    def test_gmm_known_weights_difvars_noisea(self):
        """
        Test biGMM fit with known post weights, from biGMM samples (no noise)
        1D case.
        """
        plot = False
        # pyhrf.verbose.set_verbosity(0)
        pyhrf.logger.setLevel(logging.WARNING)
        np.random.seed(2354)  # make reproducible results

        # biGMM parameters:
        mu0 = np.array([2., 5., 10.])
        v0 = 3.

        mu1 = np.array([12., 15., 20.])
        v1 = v0 / 50.

        l = .3  # weight for component 1
        v_noise = 1.5

        # generate samples from biGMM:
        nsamples = 10000
        samples = np.random.randn(nsamples) * v0 ** .5 + mu0[:, np.newaxis]
        card_c1 = int(l * nsamples)
        samples[:, :card_c1] = np.random.randn(card_c1) * v1 ** .5 + \
            mu1[:, np.newaxis]
        samples = samples.T
        noise = np.random.randn(samples.shape[0], samples.shape[1]) * v_noise
        # print 'noise shape = ', str(noise.shape)
        noisy_samples = samples + noise

        # compute sample-specific posterior weights:
        nfeats = len(mu0)
        print nfeats
        post_weights = 1 / (1 + (1 - l) / l * (v0 / v1) ** (-nfeats / 2.) *
                            np.exp(-(pm.norm2_bc(noisy_samples, mu0) / v0 -
                                     pm.norm2_bc(noisy_samples, mu1) / v1) / 2.))

        # fit GMM knowing post weights:
        mu0_est, mu1_est, v0_est, v1_est, l_est, llh = \
            pm.informedGMM_MV(samples, post_weights)

        # Some outputs:
        mus = np.array([[mu0, mu1], [mu0_est, mu1_est]]).transpose()
        vs = np.array([[v0, v1], [v0_est, v1_est]]).T
        ls = np.array([[1 - l, l], [1 - l_est, l_est]]).T

        from pyhrf.tools import get_2Dtable_string
        logger.info('means:')
        for f in xrange(samples.shape[1]):
            logger.info(get_2Dtable_string(mus[f, :, :], ['C0', 'C1'],
                                           ['true', 'estim']))
        logger.info('vars:')
        logger.info(get_2Dtable_string(vs, ['C0', 'C1'], ['true', 'estim']))
        logger.info('weights:')
        logger.info(get_2Dtable_string(ls, ['C0', 'C1'], ['true', 'estim']))

        if plot:
            import matplotlib.pyplot as plt
            from pyhrf.plot import plot_gaussian_mixture
            f = 1  # check this feature

            # plt.vlines(samples[:,f], 0, post_weights, 'r')
            # plt.vlines(samples[:,f], 0, 1-post_weights, 'b')

            plot_gaussian_mixture(np.array([[mu0[f], mu1[f]], [v0, v1]]),
                                  props=[1 - l, l], color='r')

            plot_gaussian_mixture(np.array([[mu0_est[f], mu1_est[f]],
                                            [v0_est, v1_est]]),
                                  props=[1 - l_est, l_est], color='b')

            plt.hist(samples[:, f], color='g', normed=True)
            plt.show()

        assert_array_almost_equal([mu0_est, mu1_est], [mu0, mu1], decimal=1)
        assert_array_almost_equal([v0_est, v1_est], [v0, v1], decimal=1)
        assert_array_almost_equal([1 - l_est, l_est], [1 - l, l], decimal=1)

    def test_gmm_known_weights_noisea(self):
        """
        Test biGMM fit with known post weights, from biGMM samples (no noise)
        1D case.
        """
        plot = False
        # pyhrf.verbose.set_verbosity(0)
        pyhrf.logger.setLevel(logging.WARNING)
        np.random.seed(2354)  # make reproducible results

        # biGMM parameters:
        mu0 = np.array([2., 5., 10.])
        v0 = 3.

        mu1 = np.array([12., 15., 20.])
        v1 = 3.

        l = .3  # weight for component 1
        v_noise = 1.5

        # generate samples from biGMM:
        nsamples = 10000
        samples = np.random.randn(nsamples) * v0 ** .5 + mu0[:, np.newaxis]
        card_c1 = int(l * nsamples)
        samples[:, :card_c1] = np.random.randn(card_c1) * v1 ** .5 + \
            mu1[:, np.newaxis]
        samples = samples.T
        noise = np.random.randn(samples.shape[0], samples.shape[1]) * v_noise
        # print 'noise shape = ', str(noise.shape)
        noisy_samples = samples + noise

        # compute sample-specific posterior weights:
        nfeats = len(mu0)
        print nfeats
        post_weights = 1 / (1 + (1 - l) / l * (v0 / v1) ** (-nfeats / 2.) *
                            np.exp(-(pm.norm2_bc(samples, mu0) / v0 -
                                     pm.norm2_bc(samples, mu1) / v1) / 2.))

        # fit GMM knowing post weights:
        mu0_est, mu1_est, v0_est, v1_est, l_est, llh = \
            pm.informedGMM_MV(noisy_samples, post_weights)

        # Some outputs:
        mus = np.array([[mu0, mu1], [mu0_est, mu1_est]]).transpose()
        vs = np.array([[v0, v1], [v0_est, v1_est]]).T
        ls = np.array([[1 - l, l], [1 - l_est, l_est]]).T

        from pyhrf.tools import get_2Dtable_string
        logger.info('means:')
        for f in xrange(samples.shape[1]):
            logger.info(get_2Dtable_string(mus[f, :, :], ['C0', 'C1'],
                                           ['true', 'estim']))
        logger.info('vars:')
        logger.info(get_2Dtable_string(vs, ['C0', 'C1'], ['true', 'estim']))
        logger.info('weights:')
        logger.info(get_2Dtable_string(ls, ['C0', 'C1'], ['true', 'estim']))

        if plot:
            import matplotlib.pyplot as plt
            from pyhrf.plot import plot_gaussian_mixture
            f = 1  # check this feature

            plot_gaussian_mixture(np.array([[mu0[f], mu1[f]], [v0, v1]]),
                                  props=[1 - l, l], color='r')

            plot_gaussian_mixture(np.array([[mu0_est[f], mu1_est[f]],
                                            [v0_est, v1_est]]),
                                  props=[1 - l_est, l_est], color='b')

            plt.hist(samples[:, f], color='g', normed=True)
            plt.show()

        assert_array_almost_equal([mu0_est, mu1_est], [mu0, mu1], decimal=1)
        assert_array_almost_equal([v0_est, v1_est], [v0, v1], decimal=1)
        assert_array_almost_equal([1 - l_est, l_est], [1 - l, l], decimal=1)

    def test_gmm_known_weights_noise(self):
        """
        Test biGMM fit with known post weights, from biGMM samples (no noise)
        1D case.
        """
        plot = False
        # pyhrf.verbose.set_verbosity(0)
        pyhrf.logger.setLevel(logging.WARNING)
        np.random.seed(2354)  # make reproducible results

        # biGMM parameters:
        mu0 = np.array([2., 5., 10.])
        v0 = 3.

        mu1 = np.array([12., 15., 20.])
        v1 = 3.

        l = .3  # weight for component 1
        v_noise = 1.5

        # generate samples from biGMM:
        nsamples = 10000
        samples = np.random.randn(nsamples) * v0 ** .5 + mu0[:, np.newaxis]
        card_c1 = int(l * nsamples)
        samples[:, :card_c1] = np.random.randn(card_c1) * v1 ** .5 + \
            mu1[:, np.newaxis]
        samples = samples.T
        print 'samples shape = ', str(samples.shape)
        noise = np.random.randn(samples.shape[0], samples.shape[1]) * v_noise
        print 'noise shape = ', str(noise.shape)
        noisy_samples = samples + noise

        # compute sample-specific posterior weights:
        nfeats = len(mu0)
        print nfeats
        post_weights = 1 / (1 + (1 - l) / l * (v0 / v1) ** (-nfeats / 2.) *
                            np.exp(-(pm.norm2_bc(samples, mu0) / v0 -
                                     pm.norm2_bc(samples, mu1) / v1) / 2.))

        # fit GMM knowing post weights:
        mu0_est, mu1_est, v0_est, v1_est, l_est, llh = \
            pm.informedGMM_MV(noisy_samples, post_weights)

        # Some outputs:
        mus = np.array([[mu0, mu1], [mu0_est, mu1_est]]).transpose()
        vs = np.array([[v0, v1], [v0_est, v1_est]]).T
        ls = np.array([[1 - l, l], [1 - l_est, l_est]]).T

        from pyhrf.tools import get_2Dtable_string
        logger.info('means:')
        for f in xrange(samples.shape[1]):
            logger.info(get_2Dtable_string(mus[f, :, :], ['C0', 'C1'],
                                           ['true', 'estim']))
        logger.info('vars:')
        logger.info(get_2Dtable_string(vs, ['C0', 'C1'], ['true', 'estim']))
        logger.info('weights:')
        logger.info(get_2Dtable_string(ls, ['C0', 'C1'], ['true', 'estim']))

        if plot:
            import matplotlib.pyplot as plt
            from pyhrf.plot import plot_gaussian_mixture
            f = 1  # check this feature
            plot_gaussian_mixture(np.array([[mu0[f], mu1[f]], [v0, v1]]),
                                  props=[1 - l, l], color='r')
            plot_gaussian_mixture(np.array([[mu0_est[f], mu1_est[f]],
                                            [v0_est, v1_est]]),
                                  props=[1 - l_est, l_est], color='b')
            plt.hist(samples[:, f], color='g', normed=True)
            plt.show()

        assert_array_almost_equal([mu0_est, mu1_est], [mu0, mu1], decimal=1)
        assert_array_almost_equal([v0_est, v1_est], [v0, v1], decimal=1)
        assert_array_almost_equal([1 - l_est, l_est], [1 - l, l], decimal=1)

    def test_informedGMM_parameters(self):
        """
        Check that merge is in favour of non-activ at the same feature level,
        starting from singleton clusters.
        """
        # pyhrf.verbose.set_verbosity(0)
        pyhrf.logger.setLevel(logging.WARNING)
        n_samples = 10000000

        # Generation of bi-Gaussian distribution
        mu0, v0 = 1., 0.6  # mean and standard deviation
        g0 = np.random.normal(mu0, v0, n_samples)
        l0 = 0.3
        print 'Gaussian 0: mu0 = %d, v0 = %d, lambda0 = %d' % (mu0, v0, l0)

        mu1, v1 = 5., 1.3  # mean and standard deviation
        g1 = np.random.normal(mu1, v1, n_samples)
        l1 = 0.7
        print 'Gaussian 1: mu1 = %d, v1 = %d, lambda1 = %d' % (mu1, v1, l1)

        features = g0.copy()
        features[:l1 * len(features)] = g1[:l1 * len(features)]

        alphas = 1 / (1 + (1 - l0) / l0 * (v0 / v1) ** .5 *
                      np.exp(-((features - mu0) ** 2 / v0 - (features - mu1) ** 2 / v1) / 2.))

        print 'N samples: ', n_samples
        print ''
        print 'Original parameters'
        print 'mu1:', mu1, 'mu0:', mu0
        print 'v1:', v1, 'v0:', v0
        print 'lambda:', l0
        print ''

        umu, uv, ul = pm.informedGMM_MV(features, alphas)

        print 'Updated parameters'
        print 'mu1:', umu[1], 'mu0:', umu[0]
        print 'v1:', uv[1], 'v0:', uv[0]
        print 'lambda:', ul[0]
        print ''

        assert_array_almost_equal(umu, [mu0, mu1])
        assert_array_almost_equal(uv, [v0, v1])

    def test_gmm_known_alpha0(self):
        """
        Test biGMM update with posterior weights equal to 0
        """
        plot = False
        # pyhrf.verbose.set_verbosity(0)
        pyhrf.logger.setLevel(logging.WARNING)
        np.random.seed(2354)  # make reproducible results

        # biGMM parameters:
        mu0 = 2.
        v0 = 5.
        l = 0.
        mu1 = 0.
        v1 = 0.

        # generate samples from biGMM:
        nsamples = 10000
        samples = np.random.randn(nsamples) * v0 ** .5 + mu0
        card_c1 = int(l * nsamples)
        samples[:card_c1] = np.random.randn(card_c1) * v1 ** .5 + mu1
        post_weights = np.zeros_like(samples)

        # fit GMM knowing post weights:
        mu_est, v_est, l_est = pm.informedGMM(samples, post_weights)
        assert_array_almost_equal(mu_est, [mu0, mu1], decimal=1)
        assert_array_almost_equal(v_est, [v0, v1], decimal=1)
        assert_array_almost_equal(l_est, [1 - l, l], decimal=1)

        # Some outputs:
        mus = np.array([[mu0, mu1], mu_est]).T
        vs = np.array([[v0, v1], v_est]).T
        ls = np.array([[1 - l, l], l_est]).T
        from pyhrf.tools import get_2Dtable_string
        logger.info('means:')
        logger.info(get_2Dtable_string(mus, ['C0', 'C1'], ['true', 'estim']))
        logger.info('vars:')
        logger.info(get_2Dtable_string(vs, ['C0', 'C1'], ['true', 'estim']))
        logger.info('weights:')
        logger.info(get_2Dtable_string(ls, ['C0', 'C1'], ['true', 'estim']))

        if plot:
            import matplotlib.pyplot as plt
            from pyhrf.plot import plot_gaussian_mixture
            #plt.vlines(samples, 0, post_weights, 'r')
            #plt.vlines(samples, 0, 1-post_weights, 'b')
            plot_gaussian_mixture(np.array([[mu0, mu1], [v0, v1]]),
                                  props=[1 - l, l], color='r')
            plot_gaussian_mixture(np.array([mu_est, v_est]),
                                  props=l_est, color='b')
            plt.hist(samples, color='g', normed=True)
            plt.show()

    def test_gmm_likelihood(self):
        """
        Test the log likelihood computation
        """
        # pyhrf.verbose.set_verbosity(0)
        pyhrf.logger.setLevel(logging.WARNING)
        np.random.seed(2354)  # make reproducible results

        # biGMM parameters:
        mu0 = 2.
        v0 = 5.
        mu1 = 10.
        v1 = 5.
        l = .4  # weight for component 1

        # generate samples from biGMM:
        nsamples = 10000
        samples = np.random.randn(nsamples) * v0 ** .5 + mu0
        card_c1 = int(l * nsamples)
        samples[:card_c1] = np.random.randn(card_c1) * v1 ** .5 + mu1

        # Calculate loglikelihood
        d = loglikelihood_computation(samples, mu0, v0, mu1, v1, l)
        assert_array_almost_equal(d, 0.0, decimal=1)
        # fit GMM knowing post weights:
        #mu_est, v_est, l_est = informedGMM(samples, post_weights)
        #d2 = loglikelihood_computation(samples, mu0, v0, mu1, v1, l)


class FeatureExtractionTest(unittest.TestCase):

    def setUp(self):
        # called before any unit test of the class
        self.my_param = "OK"
        self.tmp_path = pyhrf.get_tmp_path()  # create a temporary folder
        self.clean_tmp = True

    def tearDown(self):
        # called after any unit test of the class
        if self.clean_tmp:
            logger.info('clean tmp path')
            shutil.rmtree(self.tmp_path)

    def test_generate_features(self):
        # pyhrf.verbose.set_verbosity(0)
        pyhrf.logger.setLevel(logging.WARNING)
        p = np.array([1, 1, 1, 1, 1, 2, 2, 2, 2, 2], dtype=int)
        act_labels = np.array([0, 0, 1, 1, 0, 0, 0, 1, 1, 1], dtype=int)
        feat_levels = {1: (np.array([.4, .3]),  # feats non-activ for parc 1
                           np.array([5., 7.]),),  # feats activ for parc 1
                       2: (np.array([.1, .2]),  # feats non-activ for parc 2
                           np.array([9., 8.]),)}  # feats activ for parc 2

        expected_features = np.array([[.4, .3],
                                      [.4, .3],
                                      [5., 7.],
                                      [5., 7.],
                                      [.4, .3],
                                      [.1, .2],
                                      [.1, .2],
                                      [9., 8.],
                                      [9., 8.],
                                      [9., 8.]])

        features = pm.generate_features(p, act_labels, feat_levels, 0.)
        assert_array_equal(features, expected_features)

    # Test feature extraction previous to parcellation
    def test_feature_extraction(self):
        # pyhrf.verbose.set_verbosity(0)
        pyhrf.logger.setLevel(logging.WARNING)
        method = 'glm_deriv'
        data0 = simulate_fmri_data()
        dt = data0.simulation[0]['dt']
        time_length = data0.simulation[0]['duration']
        ncond = len(data0.simulation[0]['condition_defs'])
        labels = data0.simulation[0]['labels']
        territories = data0.simulation[0]['hrf_territories']
        ampl, pvalues, feats, bvars = pm.feature_extraction(
            data0, method, dt, time_length, ncond)
        assert_array_equal(feats.shape, bvars.shape)
        assert_equal(feats.shape[0], ampl.shape[0])
        fn1 = op.join(self.tmp_path, 'features_representation1.png')
        fn2 = op.join(self.tmp_path, 'features_representation2.png')
        fn3 = op.join(self.tmp_path, 'features_representation3.png')
        fn4 = op.join(self.tmp_path, 'features_representation4.png')
        pm.represent_features(
            feats, labels.T, 1. - pvalues, territories, 1, fn1)
        pm.represent_features(
            feats, labels.T, 1. - pvalues, territories, 0, fn2)
        mask = np.where(labels.flatten() == 0)
        f = feats[mask, :]
        pm.represent_features(
            f[0, :, :], labels.T[mask], 1. - pvalues[mask], territories[mask], 1, fn3)
        mask = np.where(labels.flatten() == 1)
        f = feats[mask, :]
        pm.represent_features(
            f[0, :, :], labels.T[mask], 1. - pvalues[mask], territories[mask], 1, fn4)
        print fn1
        #res = pm.represent_features(feats, labels.T, bvars[:,0])
        self.assertTrue(op.exists(fn1), msg='%s does not exist' % fn1)

        self.clean_tmp = True
    """
    # Test feature extraction previous to parcellation
    def test_feature_extraction(self):
        # pyhrf.verbose.set_verbosity(0)
        pyhrf.logger.setLevel(logging.WARNING)
        method = 'glm_deriv'
        data0 = simulate_fmri_data()
        dt = data0.simulation[0]['dt']
        time_length = data0.simulation[0]['duration']
        ncond = len(data0.simulation[0]['condition_defs'])
        labels = data0.simulation[0]['labels']
        territories = data0.simulation[0]['hrf_territories']
        ampl, feats, bvars = pm.feature_extraction(data0, method, dt, time_length, ncond)
        assert_array_equal(feats.shape,bvars.shape)
        assert_equal(feats.shape[0],ampl.shape[0])
        fn = op.join(self.tmp_path, 'features_representation.png')
        name = pm.represent_features(feats, labels.T, ampl, territories, 1, fn)
        name = pm.represent_features(feats, labels.T, ampl, territories, 0, fn)
        mask = np.where(labels.flatten()==0)
        f = feats[mask,:]
        name = pm.represent_features(f[0,:,:], labels.T[mask], ampl[mask], territories[mask], 1, fn)
        mask = np.where(labels.flatten()==1)
        f = feats[mask,:]
        name = pm.represent_features(f[0,:,:], labels.T[mask], ampl[mask], territories[mask], 1, fn)
        print fn
        #res = pm.represent_features(feats, labels.T, bvars[:,0])
        self.assertTrue(op.exists(fn), msg='%s does not exist'%fn)
        self.clean_tmp = False #HACK
"""


def simulate_fmri_data(scenario='high_snr', output_path=None):
        # Scenarios
        # low SNR level:
    if (scenario == 'low_snr'):
        m_act = 3.8          # activation magnitude
        v_noise = 1.        # noise variance
        # 1.5 normally, 0,2 no noise
    else:  # high SNR level:
        m_act = 1.8          # activation magnitude
        v_noise = 1.5       # noise variance
    v_act = .25
    v_inact = .25

    conditions_def = [Condition(name='audio', m_act=m_act, v_act=v_act,
                                v_inact=v_inact, label_map='house_sun'),
                      # Condition(name='audio', m_act= 5.8, v_act=.25,
                      #           v_inact=.25, label_map='squares_8x8'),
                      # Condition(name='video', m_act=2, v_act=.5,
                      # v_inact=.5, label_map='activated'),
                      ]

    # HRFs mapped to the 3 parcels:
    duration = 25.
    dt = .5
    primary_hrfs = [
        simu.genBezierHRF(np.arange(0, duration + dt, dt), pic=[3, 1],
                          normalize=True),
        simu.genBezierHRF(np.arange(0, duration + dt, dt), pic=[5, 1],
                          normalize=True),
        simu.genBezierHRF(np.arange(0, duration + dt, dt), pic=[7, 1],
                          normalize=True),
        simu.genBezierHRF(np.arange(0, duration + dt, dt), pic=[10, 1],
                          picw=6, ushoot=[22, -0.2], normalize=True),
    ]
    # Dictionnary mapping an item label to a quantity or a function that generates
    # this quantity
    simulation_steps = {
        'condition_defs': conditions_def,
        'nb_voxels': 400,
        # Labels
        'labels_vol': simu.create_labels_vol,  # 3D shape
        'labels': simu.flatten_labels_vol,  # 1D shape (flatten)
        # Nrls
        'nrls': simu.create_time_invariant_gaussian_nrls,
        'dt': dt,
        'duration': duration,
        'dsf': 2,
        'tr': 1.,
        # Paradigm
        'paradigm': simu.create_localizer_paradigm,
        'rastered_paradigm': simu.rasterize_paradigm,
        # HRF
        'hrf_duration': duration,
        'primary_hrfs': primary_hrfs,  # derivative of several hrfs
        'nb_hrf_territories': 4,
        #'hrf_territories_name' : '3_territories_8x8',
        'hrf_territories_name': '4_territories',
        'hrf_territories': simu.load_hrf_territories,
        # duplicate all HRF along all voxels
        'hrf': simu.create_hrf_from_territories,
        # Stim induced
        'stim_induced_signal': simu.create_stim_induced_signal,
        # Noise
        'v_gnoise': v_noise,
        'noise': simu.create_gaussian_noise,  # requires bold_shape, v_gnoise
        # Drift
        'drift_order': 4,
        'drift_var': 11.,
        'drift': simu.create_polynomial_drift,
        # Bold
        'bold_shape': simu.get_bold_shape,
        'bold': simu.create_bold_from_stim_induced,
    }
    simu_graph = Pipeline(simulation_steps)
    simu_graph.resolve()
    simu_vals = simu_graph.get_values()

    if output_path is not None:
        simu.simulation_save_vol_outputs(simu_vals, output_path)

    return FmriData.from_simulation_dict(simu_vals)


def create_features(size='2D', feat_contrast='high', noise_var=0.,
                    n_features=2):

    if size == '2D':
        true_parcellation = np.array([[1, 1, 1, 1],
                                      [1, 1, 1, 2],
                                      [1, 1, 2, 2],
                                      [2, 2, 2, 2]])
        act_clusters = np.array([[1, 1, 0, 0],
                                 [1, 1, 0, 0],
                                 [0, 0, 1, 1],
                                 [1, 0, 1, 1]])
        ker_mask = pgraph.kerMask2D_4n
    elif size == '1D':
        ker_mask = None

        true_parcellation = np.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2])
        act_clusters = np.array([0, 1, 1, 0, 0, 0, 0, 1, 0, 1, 1, 1])
    else:
        raise Exception('Unsupported size scenario: %s' % str(size))

    mask = true_parcellation > 0
    true_parcellation_flat = true_parcellation[np.where(mask)]
    act_clusters_flat = act_clusters[np.where(mask)]

    if feat_contrast == 'high':
        if n_features == 2:
            #                   non-act  act
            feat_levels = {1: ([1., 1.], [10, 17.]),
                           2: ([1., 1.], [40, 13.]), }
        elif n_features == 1:
            #                   non-act  act
            feat_levels = {1: ([1.], [10]),
                           2: ([1.], [40]), }
        else:
            raise Exception('nb features should be <= 2')

        #                  non-act  act
        act_levels = {1: ([.1], [.9]),
                      2: ([.1], [.9]), }

    elif feat_contrast == 'low':
        if n_features == 2:
            #                   non-act      act
            feat_levels = {1: ([1., 1.], [8., 7.]),
                           2: ([1., 1.], [8., 6.]), }
        elif n_features == 1:
            #                   non-act  act
            feat_levels = {1: ([1.], [8.]),
                           2: ([1.], [8.]), }
        else:
            raise Exception('nb features should be <= 2')

        #                  non-act  act
        act_levels = {1: ([.1], [.9]),
                      2: ([.1], [.9]), }

    else:
        raise Exception('Unsupported feature contrast scenario: %s'
                        % str(feat_contrast))

    features = pm.generate_features(true_parcellation_flat,
                                    act_clusters_flat,
                                    feat_levels, noise_var=noise_var)

    var = np.ones_like(features)

    act = pm.generate_features(true_parcellation_flat, act_clusters_flat,
                               act_levels).squeeze()

    logger.info('features:')
    logger.info(features.T)

    logger.info('activation levels:')
    logger.info(act.T)

    logger.info('variances:')
    logger.info(var.T)

    n_samples, n_features = features.shape

    graph = pgraph.graph_from_lattice(mask, ker_mask)

    return true_parcellation_flat, features, graph, var, act, \
        act_clusters_flat, mask


class ParcellationTest(unittest.TestCase):

    def setUp(self):
        # called before any unit test of the class
        self.my_param = "OK"
        self.tmp_path = pyhrf.get_tmp_path()  # create a temporary folder
        self.clean_tmp = True

    def tearDown(self):
        # called after any unit test of the class
        if self.clean_tmp:
            logger.info('clean tmp path')
            shutil.rmtree(self.tmp_path)

    def test_spatialward_against_ward_sk(self):
        """
        Check that pyhrf's spatial Ward parcellation is giving the same
        results as scikit's spatial Ward parcellation
        """
        # pyhrf.verbose.set_verbosity(0)
        pyhrf.logger.setLevel(logging.WARNING)
        true_parcellation, features, graph, var, act, act_labels, mask = \
            create_features("1D", "high", 0.)

        nbc = len(np.unique(true_parcellation))

        act = np.zeros_like(act) + 1.  # no activation levels
        ww = pm.spatial_ward(features, graph, nb_clusters=nbc)

        ww_sk = pm.spatial_ward_sk(features, graph, nb_clusters=nbc)

        assert_array_equal(ww.labels_, ww_sk.labels_)

    def test_spatialward_against_modelbasedspatialward(self):
        """
        Check that pyhrf's spatial Ward parcellation is giving the same
        results as scikit's spatial Ward parcellation
        """
        # pyhrf.verbose.set_verbosity(0)
        pyhrf.logger.setLevel(logging.WARNING)
        true_parcellation, features, graph, var, act, act_labels, mask = \
            create_features("2D", "high", 0.)

        nbc = len(np.unique(true_parcellation))

        act = np.zeros_like(act) + 1.  # no activation levels
        ww = pm.spatial_ward(features, graph, nb_clusters=nbc)

        ww_mb = pm.spatial_ward_with_uncertainty(features, graph, var, act,
                                                 nb_clusters=nbc,
                                                 dist_type='uward2')
        p0 = pm.align_parcellation(ww.labels_, ww_mb.labels_)
        assert_array_equal(ww.labels_, p0)

    def test_uspatialward_formula(self):
        """
        Check that pyhrf's Uncertain spatial Ward parcellation is giving the same
        results as Uncertain spatial Ward parcellation modified formula
        """
        # pyhrf.verbose.set_verbosity(0)
        pyhrf.logger.setLevel(logging.WARNING)
        true_parcellation, features, graph, var, act, act_labels, mask = \
            create_features("1D", "high", 0.)

        nbc = len(np.unique(true_parcellation))

        act = np.zeros_like(act) + 1.  # no activation levels

        ww = pm._with_uncertainty(features, graph, var, act, nb_clusters=nbc,
                                  dist_type='uward')

        ww_formula = pm._with_uncertainty(features, graph, var, act,
                                          nb_clusters=nbc, dist_type='uward2')

        assert_array_equal(ww.labels_, ww_formula.labels_)

    def save_parcellation_outputs(self, pobj, mask):
        pm.ward_tree_save(pobj, self.tmp_path, mask)

    def test_spatialward_from_forged_features(self):
        """
        Test spatial Ward on forged features
        """
        # pyhrf.verbose.set_verbosity(0)
        pyhrf.logger.setLevel(logging.WARNING)
        self.tmp_path = './'  # hack
        true_parcellation, features, graph, var, act, act_labels, mask = \
            create_features("2D", "high", 10.)

        nbc = len(np.unique(true_parcellation))

        act = np.zeros_like(act) + 1.  # no activation levels
        ww = pm.spatial_ward(features, graph, nb_clusters=nbc)

        if 1:
            self.save_parcellation_outputs(ww, mask)
            tp = expand_array_in_mask(true_parcellation, mask)
            fn_tp = op.join(self.tmp_path, 'true_parcellation.nii')
            xndarray(tp, axes_names=MRI3Daxes[:mask.ndim]).save(fn_tp)

            print 'pyhrf_view %s/*' % self.tmp_path
            self.clean_tmp = False  # hack

        pdist, common_parcels = parcellation_dist(ww.labels_,
                                                  true_parcellation)
        assert_array_equal(pdist, 3)  # non-regression

    def test_wpu_from_forged_features(self):
        """
        Test spatial Ward with uncertainty on forged features
        """
        # pyhrf.verbose.set_verbosity(0)
        pyhrf.logger.setLevel(logging.WARNING)
        true_parcellation, features, graph, var, act, act_labels, mask = \
            create_features("2D", "high", 0.)

        nbc = len(np.unique(true_parcellation))

        ww = pm.spatial_ward_with_uncertainty(features, graph, var, act,
                                              nb_clusters=nbc, save_history=False)

        if 1:
            self.save_parcellation_outputs(ww, mask)
            tp = expand_array_in_mask(true_parcellation, mask)
            fn_tp = op.join(self.tmp_path, 'true_parcellation.nii')
            xndarray(tp, axes_names=MRI3Daxes[:mask.ndim]).save(fn_tp)

            print 'pyhrf_view %s/*' % self.tmp_path
            self.clean_tmp = False  # hack

        pdist, common_parcels = parcellation_dist(ww.labels_,
                                                  true_parcellation)
        assert_array_equal(pdist, 3)  # non-regression

    def test_gmm_from_forged_features(self):
        """
        Test spatial Ward with uncertainty on forged features
        """
        # pyhrf.verbose.set_verbosity(0)
        pyhrf.logger.setLevel(logging.WARNING)
        true_parcellation, features, graph, var, act, act_labels, mask = \
            create_features("2D", "high", 0.)

        nbc = len(np.unique(true_parcellation))

        ww = pm.spatial_ward_with_uncertainty(features, graph, var, act,
                                              nb_clusters=nbc, save_history=False,
                                              dist_type='mixt')

        if 1:
            self.save_parcellation_outputs(ww, mask)
            tp = expand_array_in_mask(true_parcellation, mask)
            fn_tp = op.join(self.tmp_path, 'true_parcellation.nii')
            xndarray(tp, axes_names=MRI3Daxes[:mask.ndim]).save(fn_tp)

            print 'pyhrf_view %s/*' % self.tmp_path
            self.clean_tmp = False  # hack

        pdist, common_parcels = parcellation_dist(ww.labels_,
                                                  true_parcellation)
        assert_array_equal(pdist, 3)  # non-regression

    def test_parcellation_spatialWard_2(self):
        """
        Test WPU on a simple case.
        """

        # pyhrf.verbose.set_verbosity(0)
        pyhrf.logger.setLevel(logging.WARNING)
        features = np.array([[100., 100., 1., 1.],
                             [100., 100., 1., 1.],
                             [100., 100., 1., 1.]]).T
        n_samples, n_features = features.shape
        graph = pgraph.graph_from_lattice(np.ones((2, 2)), pgraph.kerMask2D_4n)

        var = np.ones_like(features)
        var_ini = np.ones_like(features)
        act = np.ones_like(features[:, 0])
        act_ini = np.ones_like(features[:, 0])
        ww = pm.spatial_ward_with_uncertainty(features, graph, var, act, var_ini,
                                              act_ini, nb_clusters=2)
        p = ww.labels_
        assert_array_equal(p, np.array([1, 1, 2, 2]))

    def test_parcellation_spatialWard_act_level_1D(self):
        """
        Test the ability of WPU to 'jump' non-activating positions (1D case).
        """
        # pyhrf.verbose.set_verbosity(0)
        pyhrf.logger.setLevel(logging.WARNING)
        old_settings = np.seterr('raise')
        true_parcellation = np.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 3, 3])
        act_labels = np.array([0, 1, 1, 0, 0, 0, 0, 1, 0, 1, 1, 1])
        #                      non-act  act
        feat_levels = {1: ([1., 5.], [10, 7.]),
                       3: ([1., 2.], [40, 3.]), }

        features = pm.generate_features(true_parcellation, act_labels,
                                        feat_levels)

        var = np.ones_like(features)

        #                  non-act  act
        act_levels = {1: ([.3], [4.]),
                      3: ([.3], [4.]), }

        act = pm.generate_features(true_parcellation, act_labels, act_levels)
        act = act.squeeze()

        logger.info('features:')
        logger.info(features.T)

        logger.info('activation levels:')
        logger.info(act)

        logger.info('variances:')
        logger.info(var.T)

        n_samples, n_features = features.shape
        graph = pgraph.graph_from_lattice(np.ones((1, n_samples)),
                                          pgraph.kerMask2D_4n)

        ww = pm.spatial_ward_with_uncertainty(features, graph, var, act,
                                              nb_clusters=2)
        p = ww.labels_
        if 0:
            fn = op.join(self.tmp_path, 'parcellation_tree.png')
            logger.info('fig parcellation tree: %s', fn)
            lab_colors = [('black', 'red')[l] for l in act_labels]
            pm.render_ward_tree(ww, fn, leave_colors=lab_colors)
            self.clean_tmp = False

        np.seterr(**old_settings)

        # tolerate 2 differing positions, correspond to 2 non-active
        # positions in between two different clusters
        pm.assert_parcellation_equal(p, true_parcellation, tol=2)

    def test_parcellation_mmp_act_level_1D(self):
        """
        Test the ability of MMP to 'jump' non-activating positions (1D case).
        """
        # pyhrf.verbose.set_verbosity(0)
        pyhrf.logger.setLevel(logging.WARNING)
        old_settings = np.seterr('raise')

        true_parcellation, features, graph, var, act, act_labels, mask = \
            create_features("1D", "high", 0.)

        ww = pm.spatial_ward_with_uncertainty(features, graph, var, act,
                                              nb_clusters=2,
                                              dist_type='mixt')
        p = ww.labels_
        if 0:
            fn = op.join(self.tmp_path, 'parcellation_tree.png')
            logger.info('fig parcellation tree: %s', fn)
            lab_colors = [('black', 'red')[l] for l in act_labels]
            pm.render_ward_tree(ww, fn, leave_colors=lab_colors)
            self.clean_tmp = False  # hack

        np.seterr(**old_settings)

        # tolerate 2 differing positions, correspond to 2 non-active
        # positions in between two different clusters
        pm.assert_parcellation_equal(p, true_parcellation, tol=2)

    def test_parcellation_spatialWard_act_level_2D(self):
        """
        Test the ability of WPU to 'jump' non-activating positions (2D case).
        """
        # pyhrf.verbose.set_verbosity(0)
        pyhrf.logger.setLevel(logging.WARNING)

        feat_contrast = "high"
        noise_var = 0.
        true_parcellation, features, graph, var, act, act_labels, mask = \
            create_features("2D", feat_contrast, noise_var)

        ww = pm.spatial_ward_with_uncertainty(features, graph, var, act,
                                              nb_clusters=2)
        p = ww.labels_
        logger.info('true parcellation:')
        logger.info(expand_array_in_mask(true_parcellation, mask))
        logger.info('WPU parcellation:')
        logger.info(expand_array_in_mask(p, mask))
        logger.info('act labels:')
        logger.info(act_labels)

        if 0:
            fn = op.join(self.tmp_path, 'parcellation_tree.png')
            logger.info('fig parcellation tree: %s', fn)
            lab_colors = [('black', 'red')[l] for l in act_labels]
            pm.render_ward_tree(ww, fn, leave_colors=lab_colors)
            self.clean_tmp = True

        # tolerate 2 differing positions, correspond to 2 non-active
        # positions in between two different clusters
        pm.assert_parcellation_equal(p, true_parcellation,
                                     tol_pos=act_labels == 0)

    def test_parcellation_mmp_act_level_2D(self):
        """
        Test the ability of MMP to 'jump' non-activating positions (2D case).
        """
        # pyhrf.verbose.set_verbosity(0)
        pyhrf.logger.setLevel(logging.WARNING)

        feat_contrast = "high"
        noise_var = 3.
        true_parcellation, features, graph, var, act, act_labels, mask = \
            create_features("2D", feat_contrast, noise_var)

        ww = pm.spatial_ward_with_uncertainty(features, graph, var, act,
                                              nb_clusters=2, dist_type='mixt')
        p = ww.labels_
        logger.info('true parcellation:')
        logger.info(expand_array_in_mask(true_parcellation, mask))
        logger.info('MMP parcellation:')
        logger.info(expand_array_in_mask(p, mask))
        logger.info('act labels:')
        logger.info(act_labels)

        if 0:
            fn = op.join(self.tmp_path, 'parcellation_tree.png')
            logger.info('fig parcellation tree: %s', fn)
            lab_colors = [('black', 'red')[l] for l in act_labels]
            pm.render_ward_tree(ww, fn, leave_colors=lab_colors)
            self.clean_tmp = True

        # tolerate 2 differing positions, correspond to 2 non-active
        # positions in between two different clusters
        pm.assert_parcellation_equal(p, true_parcellation,
                                     tol_pos=act_labels == 0)

    def test_parcellation_spatialWard_variance_1D(self):
        """
        Test the ability of WPU to 'jump' non-activating positions (1D case).
        """
        # pyhrf.verbose.set_verbosity(0)
        pyhrf.logger.setLevel(logging.WARNING)
        old_settings = np.seterr('raise')
        true_parcellation = np.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 3, 3])
        act_labels = np.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1])

        n = 0.5
        var = (np.random.randn(*true_parcellation.shape))[:, np.newaxis] * n
        features = (true_parcellation + var)
        act = act_labels.squeeze()

        logger.info('features:')
        logger.info(features.T)

        logger.info('activation levels:')
        logger.info(act)

        logger.info('variances:')
        logger.info(var.T)

        n_samples, n_features = features.shape
        graph = pgraph.graph_from_lattice(np.ones((1, n_samples)),
                                          pgraph.kerMask2D_4n)

        ww = pm.spatial_ward_with_uncertainty(features, graph, var, act,
                                              nb_clusters=2)
        p = ww.labels_

        np.seterr(**old_settings)

        # tolerate 2 differing positions, correspond to 2 non-active
        # positions in between two different clusters
        pm.assert_parcellation_equal(p, true_parcellation, tol=2)

    def test_parcellation_spatialWard_variance_2D(self):
        """
        Test the sensibility to variance (2D case).
        """
        # pyhrf.verbose.set_verbosity(0)
        pyhrf.logger.setLevel(logging.WARNING)

        true_parcellation = np.array([[1, 1, 1, 1],
                                      [1, 1, 1, 2],
                                      [1, 1, 2, 2],
                                      [2, 2, 2, 2]])
        var = np.array([[0.1, -0.1, -0.1, -0.1],
                        [0.1, 0.1, -0.1, -0.1],
                        [0.1, 0.1, 0.1, -0.1],
                        [0.1, 0.1, 0.1, 0.1]])

        act_clusters = np.array([[1, 1, 1, 1],
                                 [1, 1, 1, 1],
                                 [1, 1, 1, 1],
                                 [1, 1, 1, 1]])

        mask = true_parcellation > 0
        true_parcellation_flat = true_parcellation[np.where(mask)]
        act_clusters_flat = act_clusters[np.where(mask)]
        features = (true_parcellation + var).flatten()
        act = np.ones_like(act_clusters)

        logger.info('features %s:', str(features.shape))
        logger.info(features.T)

        logger.info('activation levels:')
        logger.info(act.T)

        logger.info('variances :')
        logger.info(var.T)

        n_samples, n_features = features[:, np.newaxis].shape
        graph = pgraph.graph_from_lattice(mask, pgraph.kerMask2D_4n)
        if 0:
            print graph.shape
            print features.shape
            print var.flatten().shape
            print act_clusters_flat.shape
        ww = pm.spatial_ward_with_uncertainty(features[:, np.newaxis], graph,
                                              var.flatten()[:, np.newaxis],
                                              act_clusters_flat[:, np.newaxis],
                                              nb_clusters=2)
        p = ww.labels_
        logger.info('true parcellation:')
        logger.info(true_parcellation)
        logger.info('WPU parcellation:')
        logger.info(expand_array_in_mask(p, mask))
        logger.info('act labels:')
        logger.info(act_clusters)

        if 0:
            fn = op.join(self.tmp_path, 'parcellation_tree.png')
            print 'fig parcellation tree:', fn
            lab_colors = [('black', 'red')[l] for l in act_clusters_flat]
            pm.render_ward_tree(ww, fn, leave_colors=lab_colors)
            # self.clean_tmp = False #HACK

        # tolerate 2 differing positions, correspond to 2 non-active
        # positions in between two different clusters
        pm.assert_parcellation_equal(p, true_parcellation_flat,
                                     tol_pos=act_clusters_flat == 0)

    def test_render_ward_tree(self):

        # pyhrf.verbose.set_verbosity(0)
        pyhrf.logger.setLevel(logging.WARNING)

        features = np.array([[1., 2., 4., 8., 16, 32]]).T
        act = np.array([1, 1, 1, 0, 0, 0])
        var = np.ones_like(features)

        n_samples, n_features = features.shape
        graph = pgraph.graph_from_lattice(np.ones((1, n_samples)),
                                          pgraph.kerMask2D_4n)

        n_clusters = 1
        ww = pm.spatial_ward_with_uncertainty(features, graph, var, act,
                                              nb_clusters=n_clusters)

        fn = op.join(self.tmp_path, 'parcellation_tree.png')
        logger.info('fig of parcellation tree: %s', fn)
        item_colors = [['black', 'red'][l] for l in act]
        pm.render_ward_tree(ww, fn, leave_colors=item_colors)
        self.assertTrue(op.exists(fn), msg='%s does not exist' % fn)

        # self.clean_tmp = False #HACK

    def test_ward_distance_1D_v1(self):
        # Test: inertia is high between clusters and 0 in the same cluster
        # pyhrf.verbose.set_verbosity(0)
        pyhrf.logger.setLevel(logging.WARNING)
        features = np.array([[10., 10., 10., 5., 5., 5.]]).T
        var = np.ones_like(features)
        mom_1 = np.array([1., 1., 1., 1., 1., 1.])
        c_r = np.array([1, 2, 3, 4, 5])
        c_c = np.array([0, 1, 2, 3, 4])
        var = np.ones_like(features)
        ini = np.array([0., 0., 0., 0., 0.])
        act = np.array([1., 1., 1., 1., 1., 1.])
        i1 = pm.compute_ward_dist(mom_1, features, c_r, c_c, var, act, ini)
        logger.info('inertia:')
        logger.info(i1)
        assert_equal(len(np.array(np.where(i1 > 0))), 1)

    def test_ward_distance_1D_v2(self):
        # Test effect non activation in limit between clusters
        # pyhrf.verbose.set_verbosity(0)
        pyhrf.logger.setLevel(logging.WARNING)
        features = np.array([[10.3, 10.1, 10.7, 5.1, 5.3, 5.2]]).T
        var = np.ones_like(features)
        mom_1 = np.array([1., 1., 1., 1., 1., 1.])
        c_r = np.array([1, 2, 3, 4, 5])
        c_c = np.array([0, 1, 2, 3, 4])
        var = np.ones_like(features)
        ini = np.array([0., 0., 0., 0., 0.])
        act = np.array([1., 1., 0.1, 0.1, 1., 1.])
        i1 = pm.compute_ward_dist(mom_1, features, c_r, c_c, var, act, ini)
        logger.info('inertia:')
        logger.info(i1)
        assert_equal(np.argmax(i1), 2)

    def test_ward_distance_2D(self):
        # Test
        # pyhrf.verbose.set_verbosity(0)
        pyhrf.logger.setLevel(logging.WARNING)
        true_parcellation = np.array([[1, 1, 1, 1],
                                      [1, 1, 1, 2],
                                      [1, 1, 2, 2],
                                      [2, 2, 2, 2]])
        act_clusters = np.array([[1, 1, 0, 0],
                                 [1, 1, 0, 0],
                                 [1, 1, 1, 0],
                                 [1, 1, 1, 0]])

        mask = true_parcellation > 0
        true_parcellation_flat = true_parcellation[np.where(mask)]
        act_clusters_flat = act_clusters[np.where(mask)]
        #                   non-act  act
        feat_levels = {1: ([1.], [10]),
                       2: ([3.], [40]), }
        features = pm.generate_features(true_parcellation_flat, act_clusters_flat,
                                        feat_levels, noise_var=0.)
        logger.info(features[:, 0].T)
        var = np.ones_like(features)
        #                  non-act  act
        act_levels = {1: ([.3], [4.]),
                      2: ([.3], [4.]), }
        act = pm.generate_features(true_parcellation_flat, act_clusters_flat,
                                   act_levels).squeeze()
        logger.info(act)
        mom_1 = np.ones_like(act)
        c_r = np.array([1, 2, 3, 4, 5, 5, 6, 6, 7, 7, 8, 9, 9,
                        10, 10, 11, 11, 12, 13, 13, 14, 14, 15, 15])
        c_c = np.array(
            [0, 1, 2, 0, 1, 4, 2, 5, 3, 6, 4, 5, 8, 6, 9, 7, 10, 8, 9, 12, 10, 13, 11, 14])
        ini = np.zeros((c_r.shape[0], 1))
        i1 = pm.compute_ward_dist(mom_1, features, c_r, c_c, var, act, ini)
        logger.info('inertia:')
        logger.info(i1.T)
        #assert_array_almost_equal(act_clusters(np.where(act_clusters==0)), inertia())

    def test_parcellation_spatialWard_5_sklearn(self):
        # pyhrf.verbose.set_verbosity(0)
        pyhrf.logger.setLevel(logging.WARNING)
        features0 = np.ones((25, 1))
        features0[10:] = 2
        n_samples, n_features = features0.shape
        noise = 0
        var = np.random.rand(n_samples, n_features) * noise
        var_ini = np.random.rand(n_samples, n_features) * noise
        act = np.ones_like(features0[:, 0])
        act_ini = np.ones_like(features0[:, 0])
        features = features0 + var
        graph = pgraph.graph_from_lattice(np.ones((5, 5)), pgraph.kerMask2D_4n)
        p2 = pm.spatial_ward(features, graph, nb_clusters=2)
        assert_array_equal(p2, features0.squeeze())

    def test_parcellation_spatialWard_400_nonoise(self):
        # pyhrf.verbose.set_verbosity(0)
        pyhrf.logger.setLevel(logging.WARNING)
        n_samples = 400.
        n_features = 1.
        im = np.concatenate((np.zeros(math.ceil(n_samples / 2)) + 1,
                             np.zeros(math.floor(n_samples / 2)) + 2)).reshape(n_samples, 1).astype(np.int)
        n = 0
        features = im + np.random.randn(*im.shape) * n
        graph = pgraph.graph_from_lattice(
            np.ones((20, 20)), pgraph.kerMask2D_4n)
        var = np.ones_like(features)
        var_ini = np.ones_like(features)
        act = np.ones_like(features[:, 0])
        act_ini = np.ones_like(features[:, 0])
        p0 = pm.spatial_ward(features, graph, nb_clusters=2)
        p = pm.spatial_ward_with_uncertainty(features, graph, var, act,
                                             var_ini, act_ini, nb_clusters=2)

        dist_total, common_parcels = parcellation_dist(
            p.labels_, im.squeeze() + 1)
        assert_array_equal(dist_total, 0)

    def test_hemodynamic_parcellation_wpu_2D_high_SNR(self):
        """
        test WPU on features extracted from a 2D artificial fMRI data set,
        at high SNR
        """
        # pyhrf.verbose.set_verbosity(0)
        pyhrf.logger.setLevel(logging.WARNING)
        data0 = simulate_fmri_data()
        method = 'glm_deriv'
        dt = data0.simulation[0]['dt']
        logger.info('fdata:')
        logger.info(data0)
        time_length = data0.simulation[0]['duration']
        ncond = len(data0.simulation[0]['condition_defs'])
        ampl, feats, bvars = pm.feature_extraction(data0, method, dt,
                                                   time_length, ncond)
        data0.build_graphs()
        ww = pm.spatial_ward_with_uncertainty(feats, data0.graphs[1], bvars, ampl,
                                              nb_clusters=4)
        print 'ww: '
        print ww
        p = ww.labels_
        print p
        true_parcellation_flat = data0.simulation[0]['hrf_territories']

        if len(data0.simulation[0]['labels']) > 1:
            act_clusters_flat = np.bitwise_or(
                *(a for a in data0.simulation[0]['labels']))
        else:
            act_clusters_flat = data0.simulation[0]['labels'][0]

        mask = data0.roiMask
        logger.info('true parcellation:')
        logger.info(
            expand_array_in_mask(true_parcellation_flat, mask))
        logger.info('WPU parcellation:')
        logger.info(expand_array_in_mask(p, mask))
        logger.info('act labels:')
        logger.info(expand_array_in_mask(act_clusters_flat, mask))
        pm.assert_parcellation_equal(p, true_parcellation_flat.astype(int),
                                     tol_pos=act_clusters_flat == 0)

    def test_hemodynamic_parcellation_GMM_2D_high_SNR(self):
        """
        test GMM-based parcellation on features extracted from a
        2D artificial fMRI data set, at high SNR
        """
        np.random.seed(5438)
        data0 = simulate_fmri_data('high_snr', self.tmp_path)
        # pyhrf.verbose.set_verbosity(0)
        pyhrf.logger.setLevel(logging.WARNING)

        method = 'glm_deriv'
        dt = data0.simulation[0]['dt']
        logger.debug('fdata:')
        logger.debug(data0)
        time_length = data0.simulation[0]['duration']
        ncond = len(data0.simulation[0]['condition_defs'])
        ampl, pvals, feats, bvars = pm.feature_extraction(data0, method, dt,
                                                          time_length, ncond)

        data0.build_graphs()
        ww = pm.spatial_ward_with_uncertainty(feats, data0.graphs[1], bvars,
                                              1 - pvals, nb_clusters=3,
                                              dist_type='mixt',
                                              save_history=False)

        p = ww.labels_
        true_parcellation_flat = data0.simulation[0]['hrf_territories']

        if len(data0.simulation[0]['labels']) > 1:
            labs = data0.simulation[0]['labels']
            act_clusters_flat = np.bitwise_or(*(a for a in labs))
        else:
            act_clusters_flat = data0.simulation[0]['labels'][0]

        mask = data0.roiMask
        logger.info('true parcellation:')
        logger.info(
            expand_array_in_mask(true_parcellation_flat, mask))
        logger.info('MMP parcellation:')
        logger.info(expand_array_in_mask(p, mask))
        logger.info('act labels:')
        logger.info(expand_array_in_mask(act_clusters_flat, mask))

        if 1:
            self.save_parcellation_outputs(ww, mask)
            tp = expand_array_in_mask(true_parcellation_flat, mask)
            fn_tp = op.join(self.tmp_path, 'true_parcellation.nii')
            xndarray(tp, axes_names=MRI3Daxes[:mask.ndim]).save(fn_tp)
            logger.info('pyhrf_view %s/*', self.tmp_path)
            self.clean_tmp = False  # hack

        pm.assert_parcellation_equal(p, true_parcellation_flat.astype(int),
                                     tol_pos=act_clusters_flat == 0)

    def test_parcellation_spatialWard_400_variance(self):
        # pyhrf.verbose.set_verbosity(0)
        pyhrf.logger.setLevel(logging.WARNING)
        n_samples = 400.
        n_features = 1.
        im = np.concatenate((np.zeros(math.ceil(n_samples / 2)) + 1,
                             np.zeros(math.floor(n_samples / 2)) + 2)).reshape(n_samples, 1).astype(np.int)
        n = 0.5
        var = np.random.randn(*im.shape) * n
        features = im + var
        var_ini = np.ones_like(features)
        act = np.ones_like(features[:, 0])
        act_ini = np.ones_like(features[:, 0])
        graph = pgraph.graph_from_lattice(
            np.ones((20, 20)), pgraph.kerMask2D_4n)
        p0 = pm.spatial_ward(features, graph, nb_clusters=2)
        p = pm.spatial_ward_with_uncertainty(features, graph, var, act,
                                             var_ini, act_ini, nb_clusters=2)
        from pyhrf.parcellation import parcellation_dist
        dist_total1, common_parcels = parcellation_dist(p0,
                                                        im.squeeze() + 1)
        dist_total2, common_parcels = parcellation_dist(p.labels_,
                                                        im.squeeze() + 1)
        print dist_total1
        print dist_total2
        assert_array_less(dist_total2, dist_total1)

        #p2 = pm.align_parcellation(im.squeeze(), p)
        #assert_array_equal(im.squeeze(), p2)

    def test_parcellation_history(self):

        tp, feats, g, v, act, act_labs, mask = create_features()
        nc = len(np.unique(tp))
        ww = pm.spatial_ward_with_uncertainty(feats, g, v, act,
                                              nb_clusters=nc)
        nvoxels = feats.shape[0]
        self.assertEqual(ww.history.shape, (nvoxels - nc, nvoxels))
        self.assertEqual(ww.history_choices.shape[0], nvoxels - nc)
        self.assertEqual(ww.history_choices.shape[2], nvoxels)

        #c_hist = ww.history.expand(mask, 'voxel', target_axes=MRI3Daxes)

    def test_uward_tree_save(self):
        # pyhrf.verbose.set_verbosity(0)
        pyhrf.logger.setLevel(logging.WARNING)
        tp, feats, g, v, act, act_labs, mask = create_features()
        nc = len(np.unique(tp))
        ww = pm.spatial_ward_with_uncertainty(feats, g, v, act,
                                              nb_clusters=nc)
        pm.ward_tree_save(ww, self.tmp_path, mask)
        assert_file_exists(op.join(self.tmp_path,
                                   'parcellation_uward_features.nii'))
        assert_file_exists(op.join(self.tmp_path,
                                   'parcellation_uward_variances.nii'))

        assert_file_exists(op.join(self.tmp_path,
                                   'parcellation_uward_activations.nii'))

        assert_file_exists(op.join(self.tmp_path,
                                   'parcellation_uward_history.nii'))

        assert_file_exists(op.join(self.tmp_path,
                                   'parcellation_uward_choice_history.nii'))
        assert_file_exists(op.join(self.tmp_path,
                                   'parcellation_uward_history.nii'))

        if 0:
            self.clean_tmp = False  # hack
            print 'pyhrf_view %s/*nii' % self.tmp_path

    def test_ward_tree_save(self):
        # pyhrf.verbose.set_verbosity(0)
        pyhrf.logger.setLevel(logging.WARNING)
        tp, feats, g, v, act, act_labs, mask = create_features()
        nc = len(np.unique(tp))
        ww = pm.spatial_ward_with_uncertainty(feats, g, v, act,
                                              nb_clusters=nc)
        pm.ward_tree_save(ww, self.tmp_path, mask)
        assert_file_exists(op.join(self.tmp_path,
                                   'parcellation_uward_features.nii'))
        assert_file_exists(op.join(self.tmp_path,
                                   'parcellation_uward_history.nii'))

        assert_file_exists(op.join(self.tmp_path,
                                   'parcellation_uward_choice_history.nii'))
        assert_file_exists(op.join(self.tmp_path,
                                   'parcellation_uward_history.nii'))

        if 0:
            self.clean_tmp = False  # hack
            print 'pyhrf_view %s/*nii' % self.tmp_path

    def test_mixtdist(self):
        """
        Check that merge is in favour of non-activ at the same feature level,
        starting from singleton clusters.
        """
        # pyhrf.verbose.set_verbosity(0)
        pyhrf.logger.setLevel(logging.WARNING)

        features = np.array([[1.1, 9, 1.1]]).T
        alphas = np.array([.1, .9, .9])

        c0 = np.array([1, 0, 0], dtype=int)
        c1 = np.array([0, 1, 0], dtype=int)
        c2 = np.array([0, 0, 1], dtype=int)
        cmasks = [c0, c1, c2]

        # merge 0,1
        dist_0_1 = pm.compute_mixt_dist(features, alphas, np.array([0]),
                                        np.array([1]), cmasks, [None])[0]
        logger.info('merge dist(0,1): %f', dist_0_1)

        # merge 1,2
        c1 = np.array([0, 1, 0], dtype=int)
        c2 = np.array([0, 0, 1], dtype=int)
        dist_1_2 = pm.compute_mixt_dist(features, alphas, np.array([1]),
                                        np.array([2]), cmasks, [None])[0]
        logger.info('merge dist(1,2): %f', dist_1_2)

        assert_array_less(dist_0_1, dist_1_2)



# -*- coding: utf-8 -*-

import os
import os.path as op
import cPickle
import shutil
import subprocess
import logging

from os.path import basename, splitext
from tempfile import mkdtemp

try:
    from soma_workflow.client import Job, FileTransfer
    from soma_workflow.client import Group
    from soma_workflow.client import Helper
    from soma_workflow.client import Workflow, WorkflowController
except ImportError:
    pass

import pyhrf

from pyhrf.tools._io import remote_copy
from pyhrf import xmliobak
from pyhrf.tools.cpus import available_cpu_count


logger = logging.getLogger(__name__)


class RemoteException(Exception):
    pass


def save_treatment(t, f):
    sXml = xmliobak.to_xml(t, handler=xmliobak.xmlnumpy.NumpyXMLHandler())
    fOut = open(f, 'w')
    fOut.write(sXml)
    fOut.close()


def prepare_treatment_jobs(treatment, tmp_local_dir, local_result_path,
                           local_user, local_host, remote_host, remote_user,
                           remote_path, label_for_cluster):
    """Prepare soma-workflow jobs to perform one treatment (i.e., one subject).

    Parameters
    ----------
    treatment : FMRITreatment
        the treatment defining the analysis
    tmp_local_dir : str
        a path where to store the temporary config file before sending it to the remote host
    local_result_path : str
        path where to store the final result
    local_user : str
        the user on the local host who enables SHH connection from the remote cluster
    local_host : str
        local host (used to send back the result)
    remote_host : str
        remote machine where the treatment will be run
    remote_user : str
        user login on the remote machine.
    remote_path : str
        path on the remote machine where to store ROI data and analysis results
    label_for_cluster : str
        label prefix to name job in soma-workflow

    Returns
    -------
    a tuple (job_split, jobs, dependencies, mainGroup)
    job_split (Job)
        job handling splitting of input data into ROI data
    jobs (list of Job)
        all jobs except the splitting jobs -> roi analyses, result merge, scp of result back to local host, data
        cleaning
    dependencies (list of job pairs)
        define the pipeline structure
    mainGroup (Group)
        top-level object gathering all jobs for this treatment.

    """

    # roiFiles contains the list of files that will be produced by job_split
    roiFiles, roiIds = treatment.dump_roi_datasets(dry=True)

    logger.info('Get list of splitted data files ... %d files', len(roiFiles))
    datafiles = treatment.get_data_files()

    # Make all path be relative in the treatment config file
    # so that data file can be found on the cluster file system
    treatment.replace_data_dir('./')
    remote_cfg_file = op.join(tmp_local_dir, './detectestim_remote.xml')
    treatment.set_init_param('make_outputs', False)
    logger.info('Save remote treatment to %s', remote_cfg_file)
    save_treatment(treatment, remote_cfg_file)

    logger.info('Upload input data')
    # All data which are the inputs of the workflow:
    data_to_upload = datafiles + [remote_cfg_file]
    remote_input_files = remote_copy(data_to_upload, remote_host,
                                     remote_user, remote_path)
    logger.info('Remove tmp remote cfg file')
    os.remove(remote_cfg_file)

    logger.info('Prepare jobs ...')
    logger.info('Job split ...')
    verbose_level = logger.getEffectiveLevel()
    cmd = ["pyhrf_split_roidata", "-c", basename(remote_cfg_file),
           "-v %d" % verbose_level, "-d", "./"]
    logger.info('-> %s', cmd)
    job_split = Job(cmd, working_directory=remote_path, name="roi_split")

    logger.info('Jobs JDE ...')
    jobs_jde = [Job(["pyhrf_jde_estim", "-c", basename(remote_cfg_file),
                     "-r", basename(roiFile), "-v %d" % verbose_level],
                    working_directory=remote_path,
                    name="jde_r%04d" % roiId)
                for roiFile, roiId in zip(roiFiles, roiIds)]
    logger.info('First jde job -> %s', jobs_jde[0].command)
    # Files produced by all JDE jobs, which will be then used as input of the
    # merge job:
    resultFiles = ["result_%04d.pck" % iroi for iroi in roiIds]

    logger.info('Job pack result ...')
    # Output of the merge job, which has to transfered back to local:
    remote_resultFile = './result.pck'
    logger.info('Remote result file: %s', remote_resultFile)

    cmd = ["pyhrf_pack_results", '-v1', '-o', remote_resultFile] + resultFiles
    logger.info('cmd pack result: %s', cmd)
    job_merge = Job(cmd, working_directory=remote_path,
                    name="merge_results")

    # Retrieve result file:
    # local_host = "132.166.200.5" #HACK
    # cmd = ["pyhrf_shell_cmd", "scp","-C",remote_resultFile, "%s@%s:\"%s\"" \
    #%(local_user,local_host,local_result_path)]
    cmd = ["scp", "-C", remote_resultFile, "%s@%s:\"%s\""
           % (local_user, local_host, local_result_path)]

    logger.info('cmd scp result: %s', cmd)
    job_scp_result = Job(cmd, working_directory=remote_path, name="scp_result")

    # Clean everything:
    # -> all input files, splitted roi data, result for each data, merged result:
    cmd = ["rm", "-f", remote_resultFile] + \
        map(basename, roiFiles) + resultFiles + remote_input_files
    logger.info('cmd clean: %s', cmd)
    job_clean = Job(cmd, working_directory=remote_path, name="clean_files")

    logger.info('Setup of work flow ...')

    # Build the Job lists, dependencies and group
    clean = True
    if clean:
        nodes = [job_merge, job_scp_result, job_clean] + jobs_jde
    else:
        nodes = [job_merge, job_scp_result] + jobs_jde
    dependencies = []
    for jj in jobs_jde:
        dependencies.append((job_split, jj))
        dependencies.append((jj, job_merge))
    dependencies.append((job_merge, job_scp_result))
    if clean:
        dependencies.append((job_scp_result, job_clean))

    jjGroup = Group(elements=jobs_jde, name=label_for_cluster + '-roi_jobs')
    if clean:
        elements = [job_split, jjGroup, job_merge,
                    job_scp_result, job_clean]
    else:
        elements = [job_split, jjGroup, job_merge,
                    job_scp_result]
    mainGroup = Group(name=label_for_cluster,
                      elements=elements)

    return job_split, nodes, dependencies, mainGroup


def run_soma_workflow(treatments, exec_cmd, tmp_local_dirs, server_id,
                      remote_host, remote_user, remote_pathes,
                      local_result_pathes, label_for_cluster,
                      wait_ending=False):
    """Dispatch treatments using soma-workflow.

    Parameters
    ----------
    treatments
        it is a dict mapping a treatment name to a treatment object
    exec_cmd
        it is the command to run on each ROI data.
    tmp_local_dirs
        it is a dict mapping a treatment name to a local tmp dir (used to store a temporary configuration file)
    server_id
        it is the server ID as expected by WorkflowController
    remote_host
        it is the remote machine where treatments are treated in parallel
    remote_user
        it is used to log in remote_host
    remote_pathes
        it is a dict mapping a treatment name to an existing remote dir which will be used to store ROI data and result
        files
    local_result_pathes
        it is a dict mapping a treatment name to a local path where final results will be sorted (host will send it
        there by scp)
    label_for_cluster
        it is the base name used to label workflows and sub jobs
    """

    import getpass
    from socket import gethostname

    local_user = getpass.getuser()
    local_host = gethostname()

    all_nodes = []
    all_deps = []
    all_groups = []
    split_jobs = []
    for t_id, treatment in treatments.iteritems():

        tmp_local_dir = tmp_local_dirs[t_id]
        remote_path = remote_pathes[t_id]
        local_result_path = local_result_pathes[t_id]

        sj, n, d, g = prepare_treatment_jobs(treatment, tmp_local_dir,
                                             local_result_path,
                                             local_user, local_host,
                                             remote_host,
                                             remote_user, remote_path,
                                             label_for_cluster + '-' + str(t_id))
        all_nodes.extend(n)
        all_deps.extend(d)
        all_groups.append(g)
        split_jobs.append(sj)

    # Jobs for data splitting should be done sequentially.
    # If they're done in parallel, they may flood the remote file system
    for isj in xrange(len(split_jobs)):
        if isj + 1 < len(split_jobs):
            all_deps.append((split_jobs[isj], split_jobs[isj + 1]))

    # # Be sure that all splitting jobs are done first:
    # # Is there a better way ?
    # for n in all_nodes:
    #     for sjob in split_jobs:
    #         all_deps.append((sjob,n))
    # Does not seem to work well -> maybe to many deps ?

    workflow = Workflow(
        all_nodes + split_jobs, all_deps, root_group=all_groups)

    # f = open('/tmp/workflow.pck','w')
    # cPickle.dump(workflow, f)
    # f.close()

    logger.info('Open connection ...')
    connection = WorkflowController(server_id, remote_user)

    logger.info('Submit workflow ...')
    wf_id = connection.submit_workflow(workflow=workflow,
                                       # expiration_date="",
                                       # queue="run32",
                                       name=label_for_cluster + '-' +
                                       local_user)
    #wf = connection.workflow(wf_id)

    if wait_ending:  # wait for result
        logger.info('Wait for workflow to end and make outputs ...')
        Helper.wait_workflow(wf_id, connection)

        for t_id, local_result_path in local_result_pathes.iteritems():
            treatment = treatments[t_id]
            rfilename = treatment.result_dump_file
            if rfilename is None:
                rfilename = 'result.pck'
            local_result_file = op.join(local_result_path,
                                        op.basename(rfilename))

            if not op.exists(local_result_file):
                raise Exception('Local result does not exist "%s"'
                                % local_result_file)

        if treatment.analyser.outFile is not None:
            # return result only for last treatment ...
            print 'Load result from %s ...' % local_result_file
            if splitext(local_result_file)[1] == '.gz':
                import gzip
                fresult = gzip.open(local_result_file)
            else:
                fresult = open(local_result_file)
            results = cPickle.load(fresult)
            fresult.close()
            # print 'Make outputs ...'
            #treatment.output(results, dump=False)
            logger.info('Cleaning tmp dirs ...')
            for tmp_dir in tmp_local_dirs.itervalues():
                shutil.rmtree(tmp_dir)

            return results
    else:
        logger.info('Cleaning tmp dirs ...')
        for tmp_dir in tmp_local_dirs.itervalues():
            shutil.rmtree(tmp_dir)

        logger.info('Workflow sent, returning ...')
        return []


# argv: -c, input params file, func file, output file
cfunc_marshal = "import sys;import cPickle;p=cPickle.load(open(sys.argv[1]));"\
    "import marshal,types;ff=open(sys.argv[2]);"\
    "code=marshal.loads(ff.read());ff.close();"\
    "f=types.FunctionType(code,globals());print 'p:',p;"\
    "o=f(*p[0],**p[1]);fout=open(sys.argv[3],'w');cPickle.dump(o,fout);"\
    "fout.close();"


def dump_func(func, fn):
    import marshal
    f = open(fn, 'w')
    marshal.dump(func.func_code, f)
    f.close()

import inspect


def merge_default_kwargs(func, kwargs):
    args, _, _, defaults = inspect.getargspec(func)
    if defaults is not None:
        default_kwargs = dict(zip(args[len(args) - len(defaults):], defaults))
        default_kwargs.update(kwargs)
        return default_kwargs
    else:
        return kwargs


def remote_map_marshal(func, largs=None, lkwargs=None, mode='local'):

    if largs is None:
        if lkwargs is not None:
            largs = [[]] * len(lkwargs)
        else:
            largs = []

    if lkwargs is None:
        lkwargs = [{}] * len(largs)

    lkwargs = [merge_default_kwargs(func, kw) for kw in lkwargs]

    assert len(lkwargs) == len(largs)

    all_args = zip(largs, lkwargs)

    if mode == 'local':
        return [func(*args, **kwargs) for args, kwargs in all_args]
    elif mode == 'local_with_dumps':

        func_fn = './func.marshal'
        dump_func(func, func_fn)
        results = []
        for i, params in enumerate(all_args):
            print 'params:', params
            params_fn = 'params_%d.pck' % i
            fparams = open(params_fn, 'wb')
            cPickle.dump(params, fparams)
            fparams.close()
            output_fn = 'output_%d.pck' % i
            print 'call subprocess ...'
            subprocess.call(['python', '-c', cfunc_marshal, params_fn,
                             func_fn, output_fn])
            print 'Read outputs ...'
            fout = open(output_fn)
            results.append(cPickle.load(fout))
        return results
    elif mode == 'remote_cluster':
        # FileTransfer creation for input files
        #data_dir = './rmap_data'
        data_dir = mkdtemp(prefix="sw_rmap")
        func_fn = op.join(data_dir, 'func.marshal')
        dump_func(func, func_fn)
        func_file = FileTransfer(is_input=True,
                                 client_path=func_fn,
                                 name="func_file")

        all_jobs = []
        param_files = []
        for i, params in enumerate(all_args):
            params_fn = op.join(data_dir, 'params_%d.pck' % i)
            fparams = open(params_fn, 'wb')
            cPickle.dump(params, fparams)
            fparams.close()
            param_file = FileTransfer(is_input=True,
                                      client_path=params_fn,
                                      name='params_file_%d' % i)
            param_files.append(param_file)
            output_fn = op.join(data_dir, 'output_%d.pck' % i)
            output_file = FileTransfer(is_input=False,
                                       client_path=output_fn,
                                       name='output_file_%d' % i)
            job = Job(command=['python', '-c', cfunc, param_file, func_file,
                               output_file],
                      name="rmap, item %d" % i,
                      referenced_input_files=[func_file, param_file],
                      referenced_output_files=[output_file])
            all_jobs.append(job)

        workflow = Workflow(jobs=all_jobs, dependencies=[])
        # submit the workflow
        cfg = pyhrf.cfg['parallel-cluster']
        controller = WorkflowController(cfg['server_id'], cfg['user'])

        # controller.transfer_files(fids_to_transfer)
        wf_id = controller.submit_workflow(
            workflow=workflow, name="remote_map")

        Helper.transfer_input_files(wf_id, controller)

        Helper.wait_workflow(wf_id, controller)

        Helper.transfer_output_files(wf_id, controller)

        results = []
        for i in xrange(len(all_args)):
            fout = open(op.join(data_dir, 'output_%d.pck' % i))
            results.append(cPickle.load(fout))
            fout.close()
        return results


cfunc = "import sys;import cPickle;p=cPickle.load(open(sys.argv[1]));"\
    "import %s;"\
    "o=%s(*p[0],**p[1]);fout=open(sys.argv[2],'w');cPickle.dump(o,fout);"\
    "fout.close();"


def remote_map(func, largs=None, lkwargs=None, mode='serial'):
    """
    Execute a function in parallel on a list of arguments.

    Args:
        *func* (function): function to apply on each item.
                           **this function must be importable on the remote side**
        *largs* (list of tuple): each item in the list is a tuple
                                 containing all positional argument values of the
                                 function
        *lkwargs* (list of dict): each item in the list is a dict
                                  containing all named arguments of the
                                  function mapped to their value.

        *mode* (str): indicates how execution is distributed. Choices are:

            - "serial": single-thread loop on the local machine
            - "local" : use joblib to run tasks in parallel.
                        The number of simultaneous jobs is defined in
                        the configuration section ['parallel-local']['nb_procs']
                        see ~/.pyhrf/config.cfg
            - "remote_cluster: use somaworkflow to run tasks in parallel.
                               The connection setup has to be defined
                               in the configuration section ['parallel-cluster']
                               of ~/.pyhrf/config.cfg.
            - "local_with_dumps": testing purpose only, run each task serially as
                                  a subprocess.

    Returns:
         a list of results

    Raises:
         RemoteException if any remote task has failed

    Example:
    >>> from pyhrf.parallel import remote_map
    >>> def foo(a, b=2): \
        return a + b
    >>> remote_map(foo, [(2,),(3,)], [{'b':5}, {'b':7}])
    [7, 10]
    """
    if largs is None:
        if lkwargs is not None:
            largs = [tuple()] * len(lkwargs)
        else:
            largs = [tuple()]

    if lkwargs is None:
        lkwargs = [{}] * len(largs)

    lkwargs = [merge_default_kwargs(func, kw) for kw in lkwargs]

    assert len(lkwargs) == len(largs)

    all_args = zip(largs, lkwargs)
    # print 'all_args:', all_args

    fmodule = func.__module__
    fname = '.'.join([fmodule, func.__name__])

    if mode == 'serial':
        return [func(*args, **kwargs) for args, kwargs in all_args]
    elif mode == 'local':
        try:
            from joblib import Parallel, delayed
        except ImportError:
            raise ImportError('Can not import joblib. It is '
                              'required to enable parallel '
                              'processing on a local machine.')

        if logger.getEffectiveLevel() == logging.DEBUG:
            parallel_verb = 10
        else:
            parallel_verb = 0
        if pyhrf.cfg['parallel-local']['nb_procs']:
            n_jobs = pyhrf.cfg['parallel-local']['nb_procs']
        else:
            n_jobs = available_cpu_count()
        p = Parallel(n_jobs=n_jobs, verbose=parallel_verb)
        return p(delayed(func)(*args, **kwargs)
                 for args, kwargs in all_args)

    elif mode == 'local_with_dumps':
        results = []
        for i, params in enumerate(all_args):
            # print 'params:', params
            params_fn = 'params_%d.pck' % i
            fparams = open(params_fn, 'wb')
            cPickle.dump(params, fparams)
            fparams.close()
            output_fn = 'output_%d.pck' % i
            # print 'call subprocess ...'
            subprocess.call(['python', '-c', cfunc % (fmodule, fname),
                             params_fn, output_fn])
            # print 'Read outputs ...'
            fout = open(output_fn)
            results.append(cPickle.load(fout))
        return results
    elif mode == 'remote_cluster':
        # FileTransfer creation for input files
        #data_dir = './rmap_data'
        data_dir = mkdtemp(prefix="sw_rmap")

        all_jobs = []
        param_files = []
        for i, params in enumerate(all_args):
            params_fn = op.join(data_dir, 'params_%d.pck' % i)
            fparams = open(params_fn, 'wb')
            cPickle.dump(params, fparams)
            fparams.close()
            param_file = FileTransfer(is_input=True,
                                      client_path=params_fn,
                                      name='params_file_%d' % i)
            param_files.append(param_file)
            output_fn = op.join(data_dir, 'output_%d.pck' % i)
            output_file = FileTransfer(is_input=False,
                                       client_path=output_fn,
                                       name='output_file_%d' % i)
            job = Job(command=['pyhrf_exec_pyfunc', fmodule, fname,
                               param_file, output_file],
                      name="rmap, item %d" % i,
                      referenced_input_files=[param_file],
                      referenced_output_files=[output_file])
            all_jobs.append(job)

        workflow = Workflow(jobs=all_jobs, dependencies=[])
        # submit the workflow
        cfg = pyhrf.cfg['parallel-cluster']
        controller = WorkflowController(cfg['server_id'], cfg['user'])
        # controller.transfer_files(fids_to_transfer)
        wf_id = controller.submit_workflow(
            workflow=workflow, name="remote_map")

        Helper.transfer_input_files(wf_id, controller)

        Helper.wait_workflow(wf_id, controller)

        Helper.transfer_output_files(wf_id, controller)

        results = []
        for i in xrange(len(all_args)):
            fnout = op.join(data_dir, 'output_%d.pck' % i)
            fout = open(fnout)
            o = cPickle.load(fout)
            print 'file cPickle loaded:', o
            fout.close()
            os.remove(fnout)
            if isinstance(o, Exception):
                raise RemoteException('Task %d failed' % i, o)
                if o.errno != 17:
                    raise RemoteException('Task %d failed' % i, o)
            results.append(o)
        return results



# -*- coding: utf-8 -*-

import numpy as np
import scipy.stats

from scipy.linalg import toeplitz
from scipy import diff

from pyhrf.tools import resampleToGrid

hCano = np.array([0. , 0.00025487, 0.00447602, 0.01865396, 0.04314079,
                     0.07225387, 0.0986708 , 0.11704072, 0.12522419, 0.12381613,
                     0.11500752, 0.10150135, 0.08578018, 0.0697497 , 0.05466205,
                     0.0412003 , 0.02962771, 0.01994009, 0.01199021, 0.00557441,
                     0.00048408,-0.00346985,-0.00644989,-0.00859454,-0.01002438,
                     -0.01084815,-0.01116695,-0.01107628,-0.0106662 ,-0.01002018,
                     -0.0092136 ,-0.00831224,-0.00737127,-0.00643486,-0.00553647,
                     -0.00469968,-0.00393938,-0.00326324,-0.00267318,-0.0021669 ,
                     -0.00173909,-0.00138262, 0.])
dtHCano = 0.6
tAxisHCano = np.arange(0., len(hCano)*dtHCano, dtHCano)

def genPriorCov(zc, pprcov, dt):
    #TODO : comment
    if zc :
        a = np.array([-4,6,-4,1], dtype=float)
        b = np.array([5,-4,1], dtype=float)
        matQ = np.zeros((pprcov,pprcov), dtype=float)
        matQ[0,0:3] = b
        matQ[pprcov-1,:] = [matQ[0,i] for i in range(len(matQ)-1,-1,-1)]
        matQ[1,0:4] = a
        matQ[pprcov-2,:] = [matQ[1,i] for i in range(len(matQ)-1,-1,-1)]
        c = np.concatenate(([1],np.zeros(pprcov-5)))
        r0 = np.concatenate(([1], a))
        r1 = np.zeros(pprcov-5)
        r = np.concatenate((r0,r1))
        matQ[2:pprcov-2,:] = toeplitz(c,r)
        matQ = np.divide(matQ,dt**4)
    else :
        edgePrior = np.zeros((pprcov, pprcov), dtype=float)
        edgePrior[0,0] = 1
        edgePrior[pprcov-1, pprcov-1] = 1
        b = np.array([1,-1], dtype=float)
        c = np.concatenate(([1],np.zeros((pprcov-2))))
        r1 = np.zeros(pprcov-2)
        r = np.concatenate((b,r1))
        varD1 = toeplitz(c,r)
        gradPrior = np.dot(varD1.transpose(),varD1)

        edgeGradPrior = np.zeros((pprcov, pprcov), dtype=float)
        edgeGradPrior[(0,pprcov-1),:] = gradPrior[(1,pprcov-1),:]
        epsilon = 0; #1e-8
        a = np.array([1,-2,1],dtype=float)
        c = np.concatenate(([1], np.zeros(pprcov-3)))
        r = np.concatenate((a , np.zeros(pprcov-3)))
        varD2 = toeplitz(c,r)

        matQ = edgePrior + edgeGradPrior + np.dot(varD2.transpose(),varD2)
        matQ = matQ + epsilon*np.eye(pprcov)
        matQ = np.divide(matQ, dt**4)

    return matQ


def buildFiniteDiffMatrix(order, size):

    o = order
    a = np.diff(np.concatenate((np.zeros(o),[1],np.zeros(o))),n=o)
    b = a[len(a)/2:]
    diffMat = toeplitz(np.concatenate((b, np.zeros(size-len(b)))))
    #print 'diffMat :'
    #print diffMat
    return diffMat



def genGaussianSmoothHRF(zc, length, eventdt, rh, order=2):

    prcov = length-(2*zc)
    matQ = buildFiniteDiffMatrix(order, prcov)
    matQ = np.divide(np.dot(matQ.transpose(),matQ), eventdt**(order**2))
    matL = np.array(np.transpose(np.linalg.cholesky(matQ/rh)))

    hrf = np.linalg.solve(matL, np.random.randn(prcov,1))
    #hrf = np.sqrt(rh)*hrf
    if zc :
        hrf = np.concatenate(([[0]],hrf,[[0]]))

    return (hrf[:,0],matQ)


def genExpHRF(timeAxis=np.arange(0,25,0.5), ttp=6, pa=1, pw=0.2,
              ttu=11, ua=0.2, uw=0.01):
    pic = pa*np.exp(-pw*(timeAxis-ttp)**2)
    uShoot = ua*np.exp(-uw*(timeAxis-ttu)**2)
    # calculate value at origin :
    valAtOrig = pa*np.exp(-pw*ttp**2)-ua*np.exp(-uw*ttu**2)
    # correct origin value to be zero :
    return pic-uShoot-valAtOrig


def getCanoHRF(duration=25, dt=.6, hrf_from_spm=True, delay_of_response=6.,
               delay_of_undershoot=16., dispersion_of_response=1.,
               dispersion_of_undershoot=1., ratio_resp_under=6., delay=0.):
    """Compute the canonical HRF.

    Parameters
    ----------
    duration : int or float, optional
        time lenght of the HRF in seconds
    dt : float, optional
        time resolution of the HRF in seconds
    hrf_from_spm : bool, optional
        if True, use the SPM formula to compute the HRF, if False, use the hard
        coded values and resample if necessary. It is strongly advised to use
        True.
    delay_of_response : float, optional
        delay of the first peak response in seconds
    delay_of_undershoot : float, optional
        delay of the second undershoot peak in seconds
    dispersion_of_response : float, optional
    dispersion_of_undershoot : float, optional
    ratio_resp_under : float, optional
        ratio between the response peak and the undershoot peak
    delay : float, optional
        delay of the HRF

    Returns
    -------
    time_axis : ndarray, shape (round(duration/dt)+1,)
        time axis of the HRF in seconds
    HRF : ndarray, shape (round(duration/dt)+1,)

    """

    time_axis = np.arange(0, duration+dt, dt)
    axis = np.arange(len(time_axis))

    if hrf_from_spm:
        hrf_cano = (scipy.stats.gamma.pdf(axis - delay/dt,
                                          delay_of_response/dispersion_of_response, 0,
                                          dispersion_of_response/dt) -
                    scipy.stats.gamma.pdf(axis - delay/dt,
                                          delay_of_undershoot/dispersion_of_undershoot,
                                          0, dispersion_of_undershoot/dt)/ratio_resp_under)
        hrf_cano[-1] = 0
        hrf_cano /= np.linalg.norm(hrf_cano)
    else:
        hrf_cano = resampleToGrid(time_axisHCano, hCano.copy(), time_axis)
        hrf_cano[-1] = 0.
        assert len(hrf_cano) == len(time_axis)
        hrf_cano /= (hrf_cano**2).sum()**.5

    return time_axis, hrf_cano


def getCanoHRF_tderivative(duration=25., dt=.5):

    hcano = getCanoHRF()
    hcano[-1] = 0.
    tAxis = np.arange(0, duration+dt, dt)
    hderiv = diff(hcano[1], n=1)/dt
    hderiv = np.hstack(([0], hderiv))
    return tAxis, hderiv

def genCanoBezierHRF(duration=25., dt=.6, normalize=False):

    tAxis = np.arange(0,duration+dt, dt)
    tPic = 5
    if tPic >= duration:
        tPic = duration * .3
    tus = np.round(tPic+(duration-tPic)*0.6)

    # for 15sec:
    #pic : [5, 1]
    #picw: 2
    #ushoot: [11.0, -0.20000000000000001]
    #ushootw: 3
    h = genBezierHRF(timeAxis=tAxis, pic=[tPic,1], ushoot=[tus, -0.2],
                     normalize=normalize)
    assert len(h) == len(tAxis)
    return h


def genBezierHRF(timeAxis=np.arange(0,25.5,0.6), pic=[6,1], picw=2,
                 ushoot=[15,-0.2], ushootw=3, normalize=False):
##    print 'genBezierHRF ...'
##    print 'timeAxis :', timeAxis
##    print 'pic :', pic
##    print 'picw:', picw
##    print 'ushoot:', ushoot
##    print 'ushootw:', ushootw
    prec = (timeAxis[1]-timeAxis[0])/20.
    partToPic = bezierCurve([0.,0.], [2.,0.], pic, [pic[0]-picw,pic[1]], prec)
    partToUShoot = bezierCurve(pic, [pic[0]+picw,pic[1]], ushoot,
                               [ushoot[0]-ushootw, ushoot[1]], prec)

    partToEnd = bezierCurve(ushoot, [ushoot[0]+ushootw, ushoot[1]],
                            [timeAxis[-1],0.], [timeAxis[-1]-1,0.], prec)
    hrf = range(2)
    hrf[0] = partToPic[0]+partToUShoot[0]+partToEnd[0]
    hrf[1] = partToPic[1]+partToUShoot[1]+partToEnd[1]

    # check if bezier parameters are well set to generate an injective curve
    #print 'hrf[0]:', hrf[0]
    #print 'diff :', np.diff(np.array(hrf[0]))
    assert (np.diff(np.array(hrf[0]))>=0).all()

    #return hrf
    # Resample time axis :
    iSrc = 0
    resampledHRF = [[],[]]
    for itrgT in xrange(0,len(timeAxis)):
        t = timeAxis[itrgT]
        while (iSrc+1 < len(hrf[0])) and      \
                  (np.abs(t-hrf[0][iSrc]) > np.abs(t-hrf[0][iSrc+1])):
            iSrc += 1

        resampledHRF[0].append(t)
        resampledHRF[1].append(hrf[1][iSrc])
        iSrc += 1

    resampledHRF[0] = np.array(resampledHRF[0])
    hvals = np.array(resampledHRF[1])
    if normalize:
        resampledHRF[1] = hvals / (hvals**2).sum()**.5
    return resampledHRF

def bezierCurve(p1, pc1, p2, pc2, xPrecision):

    bezierCtrlPoints = []
    bezierPoints = []

    bezierCtrlPoints.append(pc1)
    bezierCtrlPoints.append(pc2)
    bezierPoints.append(p1)
    bezierPoints.append(p2)

    precisionCriterion = 0
    nbPoints = 1

    while( precisionCriterion != nbPoints):
        precisionCriterion = 0
        nbPoints = len(bezierPoints)-1
        ip = 0
        ipC = 0
        nbP = 0
        #print 'new pass ...'
        #print 'nbPoints = ', nbPoints
        while(nbP<nbPoints) :
            e1 = bezierPoints[ip]
            e2 = bezierPoints[ip+1]
            c1 = bezierCtrlPoints[ipC]
            c2 = bezierCtrlPoints[ipC+1]

            div2 = [2.,2.]
            m1 = np.divide(np.add(c1,e1),div2)
            m2 = np.divide(np.add(c2,e2),div2)
            m3 = np.divide(np.add(c1,c2),div2)
            m4 = np.divide(np.add(m1,m3),div2)
            m5 = np.divide(np.add(m2,m3),div2)
            m = np.divide(np.add(m4,m5),div2)

            bezierCtrlPoints[ipC] = m1
            bezierCtrlPoints.insert(ipC+1, m5)
            bezierCtrlPoints.insert(ipC+1, m4)
            bezierCtrlPoints[ipC+3] = m2

            bezierPoints.insert(ip+1, m)

            # Stop criterion :
            #print 'deltax :'
            #print m[0]-bezierPoints[ip][0]
            if abs(m[0]-bezierPoints[ip][0]) < xPrecision :
                precisionCriterion += 1
                #print 'precision met !'

            nbP += 1
            ip += 2
            ipC += 4

    return ([p[0] for p in bezierPoints],[p[1] for p in bezierPoints])




# -*- coding: utf-8 -*-

import string

from collections import OrderedDict

import numpy
try:
    from scipy.io.mio import loadmat
except:
    from scipy.io.matlab import loadmat


def load_regnames(spmMatFile):
    d = loadmat(spmMatFile)
    spm = d['SPM']
    return [s[0] for s in spm['xX'][0, 0]['name'][0, 0][0]]


def loadOnsets(spmMatFile):
    d = loadmat(spmMatFile)
    spm = d['SPM']
    return get_onsets_from_spm_dict(spm)


def load_paradigm_from_mat(spmMatFile):
    d = loadmat(spmMatFile)
    spm = d['SPM']
    return (get_onsets_from_spm_dict(spm), get_tr_from_spm_dict(spm))


def load_contrasts_from_mat(spm_mat_file):
    """Returns a dictionary from an SPM.mat file of contrasts and the string
    linear combination of conditions for the corresponding contrast.

    Parameter
    ---------
    spm_mat_file : string
        path to the SPM.mat file to load contrasts from

    Returns
    -------
    contrasts : dict
        dictionary of {contrast_name: linear_condition_combination}
    """

    spm = loadmat(spm_mat_file)
    spm = spm["SPM"]

    conditions = get_conditions_list_from_mat(spm)
    contrasts_vect = get_contrasts(spm)

    return OrderedDict((clean_string_for_xml(str(contrast[0][0])),
                        contrast_vec_to_name(contrast[1][:len(conditions)], conditions))
                       for contrast in contrasts_vect if str(contrast[2][0]) == "T")


def get_conditions_list_from_mat(spm):
    """Returns a list of conditions names from an SPM.mat file with the same
    order.

    Parameter
    ---------
    spm : the SPM field of SPM structure from loadmat

    Returns
    -------
    conditions : list
        list of SPM.mat ordered conditions names
    """

    return [str(condition[0][0][0]) for condition in
                  get_field(get_field(get_field(spm, 'Sess').item()[0][0], "U"),
                            "name")[0]]


def contrast_vec_to_name(contrast_vector, conditions):
    """Returns a string of linear combination of `conditions` strings in respect
    of `contrast_vector` factors.

    Parameters
    ----------
    contrast_vector : array like
        this is a vector of factors of size `len(conditions)`
    conditions : list
        list of conditions names

    Returns
    -------
    linear_combination : string
    """

    return "".join([((str(numpy.sign(factor))[0:-1]+ "+" * bool(factor>0) + conditions[i])
                     * abs(factor)) for i, factor in
                    enumerate(contrast_vector)])


def clean_string_for_xml(name):
    """Return an equivalent of name without any other characters than
    alphanumerics ones, underscores and dashes.

    Parameter
    --------
    name : str

    Returns
    -------
    cleaned_name : str
    """

    return "".join(c for c in "_".join(name.split()) if c in string.letters + string.digits +"-_")


def load_scalefactor_from_mat(spmMatFile):
    d = loadmat(spmMatFile)
    spm = d['SPM']
    return float(get_field(get_field(spm, 'xGX')[0][0], 'gSF')[0][0][0][0])


def get_field(a, f):
    if isinstance(a, (numpy.void, numpy.ndarray)):
        return a[f]
    else:
        return getattr(a, f)


def get_tr_from_spm_dict(spm):
    if isinstance(spm, numpy.ndarray):
        spm = spm[0]
    if isinstance(spm, numpy.ndarray):
        spm = spm[0]
    return float(get_field(get_field(spm, 'xY')[0][0], 'RT'))


def get_contrasts(spm):

    if isinstance(spm, numpy.ndarray):
        spm = spm[0]
    if isinstance(spm, numpy.ndarray):
        spm = spm[0]

    if isinstance(spm, numpy.void):
        if len(spm['xCon']) > 0:
            cons = spm['xCon'][0]
            return [(c['name'], c['c'].squeeze(), c['STAT']) for c in cons]
        else:
            return []
    elif isinstance(spm, numpy.ndarray):
        cons = spm.xCon[0]
        return [(c.name, c.c.squeeze(), c.STAT) for c in cons]
    else:
        raise Exception("Type of input (%s) is unsupported" %
                        str(spm.__class__))


def get_onsets_from_spm_dict(spm):
    """
    Read paradigm from SPM structure loaded by loadmat from scipy.


    Args:
        - spm (dict): SPM structure loaded from an SPM.mat

    Return:
        - the paradigm (dict), such as:
            { <session> : { 'onsets': { <condition> : array of stim onsets },
                            'stimulusLength': { <condition> :
                                                     array of stim durations}
                          }
            }

    TODO: unit test
    """

    allOnsets = {}
    sessions = get_field(spm, 'Sess').item()[0]

    for iSess, sess in enumerate(sessions):
        ons = {}
        lgth = {}
        fU = get_field(sess, 'U')
        if isinstance(fU, numpy.ndarray) and len(fU.shape) == 2:
            su = fU[0]
        else:
            su = fU

        if numpy.iterable(su):
            su = su
        else:
            su = [su]
        for u in su:
            if isinstance(u, numpy.ndarray):
                u = u[0]
            if isinstance(u, numpy.ndarray):
                u = u[0]
            xBF = get_field(spm, 'xBF')
            if isinstance(xBF, numpy.ndarray):
                xBF = xBF[0]
            if isinstance(xBF, numpy.ndarray):
                xBF = xBF[0]
            if get_field(xBF, 'UNITS') == 'scans':
                tFactor = get_field(get_field(spm, 'xY'), 'RT')
            else:
                tFactor = 1.  # assume onsets in seconds
            name = get_field(u, 'name')
            if isinstance(name, numpy.ndarray):
                name = name[0]
            if isinstance(name, numpy.ndarray):
                name = name[0]
            if isinstance(name, numpy.ndarray):
                name = name[0]
            if name.startswith('sess'):
                iSess = int(name[4]) - 1
                name = name[6:]
            uons = get_field(u, 'ons')
            if len(uons.shape) == 2:
                o = uons[:, 0]
            else:
                o = uons
            udur = get_field(u, 'dur')
            if len(udur.shape) == 2:
                d = udur[:, 0]
            else:
                d = udur
            ons[str(name)] = numpy.array(o, dtype=float) * tFactor
            lgth[str(name)] = numpy.array(d, dtype=float) * tFactor
        allOnsets['session' + str(iSess + 1)] = {'onsets': ons,
                                                 'stimulusLength': lgth}
    return allOnsets


def get_onsets_from_spm_dict_child(spm):
    allOnsets = {}

    sessions = get_field(spm, 'Sess').item()[0]

    print 'sessions:', sessions
    for iSess, sess in enumerate(sessions):
        print 'sess :', dir(sess)
        ons = {}
        lgth = {}
        fU = get_field(sess[0], 'u')
        name = get_field(sess[0], 'name')

        uons = get_field(sess[0], 'ons')
        print 'uons:', uons
        if len(uons.shape) == 2:
            o = uons[:, 0]
        else:
            o = uons
        udur = get_field(u, 'dur')
        print 'udur:', udur, udur.shape
        if len(udur.shape) == 2:
            d = udur[:, 0]
        else:
            d = udur
        xBF = get_field(spm, 'xBF')
        if isinstance(xBF, numpy.ndarray):
            xBF = xBF[0]
        if isinstance(xBF, numpy.ndarray):
            xBF = xBF[0]
        if get_field(xBF, 'UNITS') == 'scans':
            tFactor = get_field(get_field(spm, 'xY'), 'RT')
        else:
            tFactor = 1.  # assume onsets in seconds
            ons[str(name)] = numpy.array(o, dtype=float) * tFactor
            lgth[str(name)] = numpy.array(d, dtype=float) * tFactor
        allOnsets['session' + str(iSess + 1)] = {'onsets': ons,
                                                 'stimulusLength': lgth}
    return allOnsets
